import gzip
import io
import json
import pickle
import boto3
import redis
from ml_recsys_tools.utils.logger import simple_logger as logger


class RedisTable(redis.StrictRedis):

    def __init__(self, host_url, table_name, timeout=10, **kwargs):
        self.table_name = table_name
        super().__init__(
            host=host_url, port=6379, decode_responses=False,
            socket_timeout=timeout, socket_connect_timeout=timeout,
            **kwargs)

    @staticmethod
    def _encode_data(data, compress):
        if data:
            if compress:
                compress = compress if isinstance(compress, int) and 1<=compress<=9 else 1
                data = gzip.compress(json.dumps(data).encode(), compress)
            else:
                data = json.dumps(data)
        return data

    def set_json(self, key_name, key_value, data, compress=False, **kwargs):
        return super().set(
            self.table_index_key(key_name, key_value),
            self._encode_data(data, compress),
            **kwargs)

    def set_json_to_pipeline(
            self, pipeline, key_name, key_value, data, compress=False, **kwargs):
        return pipeline.set(
            self.table_index_key(key_name, key_value),
            self._encode_data(data, compress),
            **kwargs)

    def table_index_key(self, key, value):
        return self.table_name + ':' + key + ':' + value

    def query(self, index_key, index_value):
        """
        example:
            table = RedisTable('p-bla-bla')
            result_dict = table.query('uuid', '1234')

        :returns response JSON as dict, None if not found, or response string if JSON conversion fails
        """
        key = self.table_index_key(index_key, index_value)
        data = self.get(key)
        if data:
            try:
                data = gzip.decompress(data)
            except (OSError, TypeError):
                pass  # not gzip compressed
            try:
                data = json.loads(data)
            except Exception as e:
                logger.exception(e)
                logger.error('Failed unpacking redis query. key: %s, response: %s' %
                             (key, data))
        return data


class S3FileIO:
    def __init__(self, bucket_name):
        self.bucket_name = bucket_name

    def write_binary(self, data, remote_path, compress=True):
        try:
            client = boto3.client('s3')
            if compress:
                try:
                    # https://stackoverflow.com/questions/33562394/gzip-raised-overflowerror-size-does-not-fit-in-an-unsigned-int
                    data = gzip.compress(data, 1)
                except OverflowError:
                    pass
            with io.BytesIO(data) as f:
                client.upload_fileobj(
                    Fileobj=f,
                    Bucket=self.bucket_name,
                    Key=remote_path)
        except Exception as e:
            logger.error('Failed writing to S3: %s' % str(e))
            logger.exception(e)

    def read(self, remote_path):
        try:
            client = boto3.client('s3')
            ## for some reason this returns empty sometimes, but get_object works..
            # with io.BytesIO() as f:
            #     client.download_fileobj(
            #         Bucket=self.bucket_name,
            #         Key=remote_path,
            #         Fileobj=f)
            #     data = f.read()
            data = client.get_object(
                Bucket=self.bucket_name, Key=remote_path)['Body'].\
                read()
            try:
                data = gzip.decompress(data)
            except OSError:
                pass
            return data
        except Exception as e:
            logger.error('Failed reading from S3: %s' % str(e))
            logger.exception(e)

    def pickle(self, obj, remote_path):
        logger.info('S3: pickling to %s' % remote_path)
        return self.write_binary(pickle.dumps(obj), remote_path)

    def unpickle(self, remote_path):
        logger.info('S3: unpickling from %s' % remote_path)
        return pickle.loads(self.read(remote_path))

    def local_to_remote(self, local_path, remote_path):
        logger.info('S3: copying from %s to %s' % (local_path, remote_path))
        with open(local_path, 'rb') as local:
            self.write_binary(local.read(), remote_path)

    def remote_to_local(self, remote_path, local_path):
        logger.info('S3: copying from %s to %s' % (remote_path, local_path))
        with open(local_path, 'wb') as local:
            local.write(self.read(remote_path))




import logging.config
import os

import sys

LOGGER = logging.getLogger('ml-logger')
short_time_fmt = logging.Formatter('[%(asctime)s:%(levelname)s] %(message)s', datefmt='%H:%M')


def add_file_output(logger, filename, level=logging.DEBUG):
    dir_name = os.path.dirname(filename)
    if not os.path.exists(dir_name):
        os.mkdir(dir_name)
    f_handler = logging.FileHandler(filename)
    f_handler.setLevel(level)
    f_handler.setFormatter(short_time_fmt)
    logger.addHandler(f_handler)
    return logger


def console_logger():
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    handler.setFormatter(short_time_fmt)
    logger.addHandler(handler)

    return logger


class NegativeFilter(logging.Filter):
    def __init__(self, match_string):
        super().__init__()
        self._match_string = match_string

    def filter(self, record):
        return not (self._match_string in record.getMessage())


simple_logger = console_logger()



from abc import abstractmethod

import pandas as pd
import numpy as np
from lightgbm import LGBMRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.metrics import r2_score

from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF
from ml_recsys_tools.recommenders.factorization_base import BaseFactorizationRecommender
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender
from ml_recsys_tools.recommenders.recommender_base import BasePredictorRecommender


class BaseFactorsRegressor(BasePredictorRecommender):

    default_regressor_params = {}
    default_factorizer_params = {}

    factorizer_class = BaseFactorizationRecommender
    regressor_class = RandomForestRegressor  # placeholder for generic regressor

    def __init__(self,
                 stacking_split=0.5,
                 factors_prediction=True,
                 user_factors=True,
                 item_factors=True,
                 item_features=True,
                 target_transform='log',
                 regressor_params=None,
                 item_features_params=None,
                 factorizer_params=None,
                 **kwargs):
        super().__init__(**kwargs)
        self.stacking_split = stacking_split
        self.factors_prediction = factors_prediction
        self.user_factors = user_factors
        self.item_factors = item_factors
        self.item_features = item_features
        self.target_transform_func = target_transform
        self._check_param_keys_conflicts()
        self.item_features_params = item_features_params \
            if item_features_params is not None else {}
        self.regressor_params = self._dict_update(
            self.default_regressor_params, regressor_params)
        self.factorizer_params = self._dict_update(
            self.default_factorizer_params, factorizer_params)

    def _check_param_keys_conflicts(self):
        assert len(
            set(self.default_regressor_params.keys()).intersection(
            set(self.default_factorizer_params.keys()))) == 0

    def set_params(self, **params):
        params = self._pop_set_params(
            params, ['stacking_split', 'factors_prediction',
                     'user_factors', 'item_factors',
                     'target_transform', 'item_features'])
        params = self._pop_set_dict(
            self.regressor_params, params, self.default_regressor_params.keys())
        params = self._pop_set_dict(
            self.factorizer_params, params, self.default_factorizer_params.keys())
        super().set_params(**params)

    def _init_regressor(self):
        self._regressor = self.regressor_class(**self.regressor_params)

    def _init_factorizer(self):
        self._factorizer = self.factorizer_class()
        self._factorizer.set_params(**self.factorizer_params)

    def _transform_targets(self, targets):
        if self.target_transform_func == 'log':
            return np.log(1 + np.min(targets))
        elif self.target_transform_func in [None, 'none']:
            return targets
        else:
            return NotImplementedError('Uknown target transform function')

    def _make_features_df(self, df_inds, with_targets=True):
        cols = [self._uid_col, self._iid_col] + \
               ([self._rating_col] if with_targets else [])
        df_feat = df_inds[cols].copy()

        if with_targets:
            df_feat[self._rating_col] = self._transform_targets(
                df_feat[self._rating_col].values)

        if self.item_features:
            df_feat = pd.merge(
                df_feat, self.df_item_features.reset_index(),
                left_on=self._iid_col, right_on='index', how='left'). \
                drop('index', axis=1)

        if self.factors_prediction:
            df_feat[self._prediction_col] = self._factorizer._predict(
                df_feat[self._uid_col].values, df_feat[self._iid_col].values)

        if self.user_factors:
            df_feat = pd.merge(
                df_feat, self.df_user_factors.reset_index(),
                left_on=self._uid_col, right_on='index', how='left'). \
                drop('index', axis=1)

        if self.item_factors:
            df_feat = pd.merge(
                df_feat, self.df_item_factors.reset_index(),
                left_on=self._iid_col, right_on='index', how='left'). \
                drop('index', axis=1)

        df_feat.drop([self._uid_col, self._iid_col], axis=1, inplace=True)

        return df_feat

    def _df_ids_to_df_inds(self, df):
        df = self._factorizer.sparse_mat_builder.remove_unseen_labels(df)
        return self._factorizer.sparse_mat_builder.add_encoded_cols(df)

    def _make_pos_neg_feat_df(self, df):
        df_inds = self._df_ids_to_df_inds(df)

        if len(df_inds):
            df_pos = self._make_features_df(df_inds, with_targets=True)

            df_inds_neg = df_inds.assign(
                **{'iid_coord': df_inds['iid_coord'].sample(n=len(df_inds)).values,
                   'rating': 0})
            df_neg = self._make_features_df(df_inds_neg, with_targets=True)

            return pd.concat([df_pos, df_neg], axis=0)

    def _set_item_features_df(self, train_obs):
        if hasattr(train_obs, 'df_items'):
            ext_feat = train_obs.get_item_features_for_obs(**self.item_features_params)
            mat_builder = train_obs.get_sparse_matrix_helper()
            feat_mat = ext_feat.fit_transform_ids_df_to_mat(
                mat_builder.iid_encoder, mode='encode')
            self.df_item_features = pd.DataFrame(
                index=np.arange(feat_mat.shape[0]),
                data=feat_mat)
        else:
            raise ValueError('df_items not present in training ObservationHandler')

    def fit(self, train_obs: ObservationsDF, *args, **kwargs):
        factors_obs, reg_obs = train_obs.split_train_test(ratio=self.stacking_split)
        self._set_item_features_df(train_obs)
        self._set_data(factors_obs)
        self._fit_factorizer(factors_obs)
        self._fit_regressor(reg_obs)

    def _fit_factorizer(self, factors_obs, **fit_params):
        self._init_factorizer()
        self._factorizer.fit(factors_obs, **fit_params)
        self._rating_col = self.sparse_mat_builder.rating_source_col
        self._uid_col = self.sparse_mat_builder.uid_col
        self._iid_col = self.sparse_mat_builder.iid_col
        self.df_user_factors = self._factorizer.user_factors_dataframe()
        self.df_item_factors = self._factorizer.item_factors_dataframe()

    def _fit_regressor(self, reg_obs, **fit_params):
        df_train_reg = self._make_pos_neg_feat_df(reg_obs.df_obs)
        self._init_regressor()
        self._regressor.fit(df_train_reg.drop(self._rating_col, axis=1).values,
                            df_train_reg[self._rating_col].values, **fit_params)

    def _predict(self, user_inds, item_inds):
        df_inds = pd.DataFrame({
            self._uid_col: np.array(user_inds).ravel(),
            self._iid_col: np.array(item_inds).ravel()})
        df_with_feat = self._make_features_df(df_inds, with_targets=False)
        return self._regressor.predict(df_with_feat.values)

    def evaluate_regressor(self, test_dfs, test_names):
        scores = []
        for test_df in test_dfs:
            df_test_ref = self._make_pos_neg_feat_df(test_df)
            if df_test_ref is not None:
                y_true = df_test_ref[self._rating_col].values
                y_pred = self._regressor.predict(df_test_ref.drop(self._rating_col, axis=1).values)
                score = r2_score(y_true, y_pred)
            else:
                score = None
            scores.append({'r2': score})
        return pd.DataFrame(data=scores, index=test_names)

    def _get_recommendations_flat(
            self, user_ids, n_rec, item_ids=None, exclude_training=True, **kwargs):

        return self.get_recommendations_exact(
            user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
            exclude_training=exclude_training, results_format='flat')

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          remove_self=True, embeddings_mode=None,
                          simil_mode='cosine', results_format='lists'):
        raise NotImplementedError

    def predict_for_user(self, user_id, item_ids, rank_training_last=True, sort=True):
        raise NotImplementedError

    def _predict_rank(self, test_mat, train_mat=None):
        raise NotImplementedError


class BaseLFMRegRec(BaseFactorsRegressor):
    default_factorizer_params = dict(**LightFMRecommender.default_model_params,
                                     **{'epochs': 20})
    factorizer_class = LightFMRecommender


class BaseRFRegRec(BaseFactorsRegressor):
    default_regressor_params = dict(n_estimators=100, n_jobs=-1)
    regressor_class = RandomForestRegressor


class BaseLGBMRecReg(BaseFactorsRegressor):
    default_regressor_params = dict(
        boosting_type="gbdt",
        objective='regression',
        num_leaves=31,
        max_depth=-1,
        learning_rate=0.1,
        n_estimators=100,
        min_child_weight=1e-3,  # what's the right setting for this case?
        min_child_samples=20,  # what's the right setting for this case?
        subsample=1.,
        subsample_freq=1,
        colsample_bytree=1.,
        reg_alpha=0.,
        reg_lambda=0.,
        silent=True,
        verbose=-1,
    )
    regressor_class = LGBMRegressor


class RFonLFMRegRec(BaseLFMRegRec, BaseRFRegRec):
    pass


class LGBMonLFMRegRec(BaseLFMRegRec, BaseLGBMRecReg):
    pass


"""
using multiple test sets
"""

# dataset: download and prepare dataframes
import pandas as pd
from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()

# read the interactions dataframe and create a data handler object and  split to train and test
ratings_df = pd.read_csv(rating_csv_path)
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF

obs = ObservationsDF(ratings_df)
train_obs, test_obs = obs.split_train_test(ratio=0.2, users_ratio=0.2)

def construct_multiple_test_sets(test_df, train_df):
    # by user history - active and inactive users
    user_hist_counts = train_df.userid.value_counts()
    user_hist_counts.hist(bins=100, alpha=0.5)
    active_users = user_hist_counts[user_hist_counts >= 300].index.tolist()
    test_df_act_us = test_df[test_df.userid.isin(active_users)]
    test_df_nonact_us = test_df[~test_df.userid.isin(active_users)]

    # by item popularity- popular and unpopular items
    item_hist_counts = train_df.itemid.value_counts()
    item_hist_counts.hist(bins=100, alpha=0.5)
    popular_items = item_hist_counts[item_hist_counts >= 1000].index.tolist()
    test_df_pop_movies = test_df[test_df.itemid.isin(popular_items)]
    test_df_nonpop_movies = test_df[~test_df.itemid.isin(popular_items)]

    test_dfs = [test_df, test_df_act_us, test_df_nonact_us, test_df_pop_movies, test_df_nonpop_movies]
    test_names = ['all ', 'active users ', 'inactive users ', 'popular movies ', 'unpopular movies ']
    df_lens = [len(t) for t in test_dfs]
    print('Test DFs counts: ' + str(list(zip(test_names, df_lens))))
    return test_dfs, test_names

test_dfs, test_names = construct_multiple_test_sets(test_df=test_obs.df_obs, train_df=train_obs.df_obs)

# evaluation
lfm_rec = LightFMRecommender()
lfm_rec.fit(train_obs, epochs=10)
print(lfm_rec.eval_on_test_by_ranking(
    test_dfs=test_dfs, test_names=test_names, prefix='lfm regular '))



import warnings

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import copy
import scipy.sparse as sp

from multiprocessing.pool import Pool

from sklearn.model_selection import train_test_split

from ml_recsys_tools.utils.parallelism import batch_generator, N_CPUS
from ml_recsys_tools.utils.instrumentation import LogCallsTimeAndOutput
from ml_recsys_tools.utils.logger import simple_logger as logger
from ml_recsys_tools.utils.pandas_utils import console_settings
from ml_recsys_tools.utils.sklearn_extenstions import PDLabelEncoder

console_settings()

RANDOM_STATE = 42


class ObservationsDF(LogCallsTimeAndOutput):

    def __init__(self, df_obs=None, uid_col='userid', iid_col='itemid', timestamp_col=None,
                 rating_col='rating', verbose=True, **kwargs):
        super().__init__(verbose, **kwargs)
        self.df_obs = df_obs
        self.uid_col = uid_col
        self.iid_col = iid_col
        self.rating_col = rating_col
        self.timestamp_col = timestamp_col

        if self.df_obs is not None:
            self.df_obs[self.uid_col] = self.df_obs[self.uid_col].astype(str)
            self.df_obs[self.iid_col] = self.df_obs[self.iid_col].astype(str)
            self.df_obs[self.rating_col] = self.df_obs[self.rating_col].astype(float)

        self._check_duplicated_interactions()

    def __len__(self):
        return len(self.df_obs)

    def __repr__(self):
        return super().__repr__() + ', %d Observations' % len(self)

    def __add__(self, other):
        self.df_obs = pd.concat([self.df_obs, other.df_obs])
        self._check_duplicated_interactions()
        return self

    @property
    def user_ids(self):
        return self.df_obs[self.uid_col].values

    @property
    def item_ids(self):
        return self.df_obs[self.iid_col].values

    @property
    def ratings(self):
        return self.df_obs[self.rating_col].values

    @property
    def timestamps(self):
        return self.df_obs[self.timestamp_col].values

    def _check_duplicated_interactions(self):
        dups = self.df_obs.duplicated([self.uid_col, self.iid_col])
        if dups.sum():
            logger.warn('ObservationsDF: Dropping %s duplicate interactions.'
                        % str(dups.sum()))
            self.df_obs = self.df_obs[~dups]

    def _user_and_item_counts(self, plot=False):
        if len(self.df_obs):
            items_per_user = self.df_obs.groupby(self.uid_col).apply(len). \
                sort_values(ascending=False).reset_index(name='items_per_user')
            users_per_items = self.df_obs.groupby(self.iid_col).apply(len). \
                sort_values(ascending=False).reset_index(name='users_per_items')
        else:
            raise ValueError('Observations dataframe is empty')

        if plot:
            items_per_user.value_counts().reset_index(drop=True). \
                plot(logx=True, logy=True, grid=True, label='items_per_user')
            users_per_items.value_counts().reset_index(drop=True). \
                plot(logx=True, logy=True, grid=True, label='users_per_items')
            plt.ylabel('count of ..')
            plt.xlabel('with that many')
            plt.legend()
        return items_per_user, users_per_items

    def sample_observations(self,
                            n_users=None,
                            n_items=None,
                            method='random',
                            min_user_hist=0,
                            min_item_hist=0,
                            random_state=None):
        """
        :param n_users: number of users to sample
        :param n_items: number of listings to sample
        :param method: either 'random' or 'top' (sample the top users and top items by views)
        :param min_user_hist: minimal number of unique items viewed by a user
        :param min_item_hist: minimal number of unique users who viewed a item
        :return: dataframe
        """
        if method == 'top' or min_user_hist or min_item_hist:
            items_per_user, users_per_items = self._user_and_item_counts()

            if min_user_hist:
                items_per_user = items_per_user[
                    items_per_user['items_per_user'] >= min_user_hist]
            users_filt = items_per_user[self.uid_col].values

            if min_item_hist:
                users_per_items = users_per_items[
                    users_per_items['users_per_items'] >= min_item_hist]
            item_filt = users_per_items[self.iid_col].values

        else:
            users_filt = self.df_obs[self.uid_col].unique()
            item_filt = self.df_obs[self.iid_col].unique()

        if n_users is None:
            users_sample = users_filt
        elif method == 'random':
            # users_sample = users_filt.sample(n_users, random_state=random_state)
            np.random.seed(random_state)
            users_sample = np.random.choice(users_filt, n_users, replace=False)
        elif method == 'top':
            users_sample = users_filt[:n_users]
        else:
            raise ValueError('Uknown sampling method')

        if n_items is None:
            item_sample = item_filt
        elif method == 'random':
            # item_sample = item_filt.sample(n_items, random_state=random_state)
            np.random.seed(random_state)
            item_sample = np.random.choice(item_filt, n_items, replace=False)
        elif method == 'top':
            item_sample = item_filt[:n_items]
        else:
            raise ValueError('Uknown sampling method')

        sample_df = self.df_obs[(self.df_obs[self.iid_col].isin(item_sample)) &
                                (self.df_obs[self.uid_col].isin(users_sample))]

        other = copy.deepcopy(self)
        other.df_obs = sample_df.copy()

        return other

    def filter_columns_by_df(self, other_df_obs):
        """
        removes users or items that are not in the other user dataframe
        :param other_df_obs: other dataframe, that has the same structure (column names)
        :return: new observation handler
        """
        other = copy.deepcopy(self)
        other.df_obs = self.df_obs[
            (self.df_obs[self.iid_col].isin(other_df_obs[self.iid_col].unique())) &
            (self.df_obs[self.uid_col].isin(other_df_obs[self.uid_col].unique()))]
        return other

    def remove_interactions_by_df(self, other_df_obs):
        """
        removes all interactions that are present in the other dataframe.
        e.g. remove training examples from the full dataframe
        :param other_df_obs: other dataframe
        :return: new observation handler
        """
        df_filtered = pd.merge(
            self.df_obs, other_df_obs, on=[self.iid_col, self.uid_col], how='left')
        df_filtered = df_filtered[df_filtered[self.rating_col + '_y'].isnull()]. \
            rename({self.rating_col + '_x': self.rating_col}, axis=1). \
            drop(self.rating_col + '_y', axis=1)
        other = copy.deepcopy(self)
        other.df_obs = df_filtered
        return other

    def user_filtered_df(self, user):
        return self.df_obs[self.df_obs[self.uid_col] == user]

    def items_filtered_df(self, item):
        return self.df_obs[self.df_obs[self.iid_col] == item]

    @staticmethod
    def time_filter_on_df(df, time_col, days_delta_tuple):
        time_max = pd.Timestamp(df[time_col].max()) - pd.Timedelta(days=min(days_delta_tuple))
        time_delta = pd.Timedelta(days=abs(days_delta_tuple[0] - days_delta_tuple[1]))
        time_filt = (str(time_max - time_delta) < df[time_col].astype(str)) & \
                    (df[time_col].astype(str) <= str(time_max))
        return time_filt

    def get_sparse_matrix_helper(self):
        mat_builder = InteractionMatrixBuilder(
            self.df_obs, users_col=self.uid_col,
            items_col=self.iid_col, rating_col=self.rating_col)
        return mat_builder

    def split_train_test_to_dfs(self, ratio=0.2, users_ratio=1.0,
                                time_split_column=None, random_state=None):
        """
        splits the underlying dataframe  into train and test using the arguments to choose the method.
        If time_split_column is provided: the split is done using that column and provided ratio.
        Otherwise a random split either within a subset of users or not is done.

        Raises an error if both users_ratio and time_split_column are not their defaults,
        because only one method can be used.

        :param ratio: the fraction of data that should be in the resulting test segment.
        :param users_ratio: the fraction of users for which a split should be done. Using this arguments
            speeds up subsequent evaluation, because prediction will only be calculated for a subset
            of users. The ratio is only applied to those users.
        :param time_split_column: an optional argument that changes the split into a time split.
        :param random_state: a random state that is only used if the split if a shuffle split (not time).

        :return: two ObservationsDF objects, one with training data and the other with test data.
        """

        if users_ratio < 1.0 and time_split_column is not None:
            raise ValueError('Can either split by time, or for subset of users, not both.')

        if users_ratio < 1.0:
            return train_test_split_by_col(
                self.df_obs, col_ratio=users_ratio, test_ratio=ratio,
                col_name=self.uid_col, random_state=random_state)

        elif time_split_column:
            self.df_obs.sort_values(time_split_column, inplace=True)
            split_ind = int((len(self.df_obs)-1) * ratio)
            return self.df_obs.iloc[:-split_ind].copy(), self.df_obs.iloc[-split_ind:].copy()

        else:
            return train_test_split(self.df_obs, test_size=ratio, random_state=random_state)

    def split_by_time_col(self, time_col, days_delta_tuple):

        time_filt = self.time_filter_on_df(
            self.df_obs, time_col=time_col, days_delta_tuple=days_delta_tuple)

        df_train = self.df_obs[~time_filt].copy()
        df_test = self.df_obs[time_filt].copy()

        train_other = copy.deepcopy(self)
        train_other.df_obs = df_train

        test_other = copy.deepcopy(self)
        test_other.df_obs = df_test

        return train_other, test_other

    def split_train_test(self, ratio=0.2, users_ratio=1.0, time_split_column=None, random_state=None):
        """
        splits the object into train and test objects using the arguments to choose the method.
        If time_split_column is provided: the split is done using that column and provided ratio.
        Otherwise it's a random split either within a subset of users or not is done.

        :param ratio: the fraction of data that should be in the resulting test segment
        :param users_ratio: the fraction of users for which a split should be done. Using this arguments
            speeds up subsequent evaluation, because prediction will only be calculated for a subset
            of users. The ratio is only applied to those users.
        :param time_split_column: an optional argument that changes the split into a time split.
        :param random_state: a random state that is only used if the split if a shuffle split (not time).

        :return: two ObservationsDF objects, one with training data and the other with test data.
        """
        df_train, df_test = self.split_train_test_to_dfs(
            ratio=ratio,
            users_ratio=users_ratio,
            random_state=random_state,
            time_split_column=time_split_column)

        train_other = copy.deepcopy(self)
        train_other.df_obs = df_train

        test_other = copy.deepcopy(self)
        test_other.df_obs = df_test

        return train_other, test_other

    def generate_time_batches(self, n_batches, time_col, days_delta=1):
        remaining = copy.deepcopy(self)
        inf_days = 10000  # just lots of days
        for i in reversed(range(n_batches)):
            # this is done so that the earliest batch contains all
            # the data that's not in later batches
            delta = days_delta if i < (n_batches - 1) else inf_days

            remaining, batch_obs = remaining.split_by_time_col(
                time_col, (i * days_delta, i * days_delta + delta))
            yield batch_obs.df_obs


def train_test_split_by_col(df, col_ratio=0.2, test_ratio=0.2, col_name='userid', random_state=None):
    # split field unique values
    vals_train, vals_test = train_test_split(
        df[col_name].unique(), test_size=col_ratio, random_state=random_state)

    df_train_col = df[df[col_name].isin(vals_train)]
    df_test_col = df[df[col_name].isin(vals_test)]

    # split within selected field values
    df_non_test_items, df_test = train_test_split(
        df_test_col, test_size=test_ratio, random_state=random_state)

    # concat train dfs
    df_train = pd.concat([df_train_col, df_non_test_items])
    # shuffle
    df_train = df_train.sample(frac=1).reset_index(drop=True)
    return df_train, df_test


class InteractionMatrixBuilder(LogCallsTimeAndOutput):

    # this filter is due to this issue, can be removed with next version of sklearn (should be fixed)
    # https://stackoverflow.com/questions/49545947/sklearn-deprecationwarning-truth-value-of-an-array
    warnings.filterwarnings(message="The truth value of an empty array is ambiguous. "
                                    "Returning False, but in future this will result in an error. "
                                    "Use `array.size > 0` to check that an array is not empty.",
                            action='ignore', category=DeprecationWarning)

    def __init__(self, source_df, users_col='userid', items_col='adid', rating_col='rating', verbose=True):
        super().__init__(verbose=verbose)

        self.uid_source_col = users_col
        self.iid_source_col = items_col
        self.rating_source_col = rating_col
        self.uid_col = 'uuid_coord'
        self.iid_col = 'iid_coord'

        all_uids = source_df[self.uid_source_col].unique()
        all_iids = source_df[self.iid_source_col].unique()

        # shuffling because np.unique() returns elements in almost sorted order by counts,
        # and it's probably not a good thing: it changes regional sparsity,
        # and at a later stage might be sampled / iterated in order
        np.random.shuffle(all_uids)
        np.random.shuffle(all_iids)

        self.n_rows = len(all_uids)
        self.n_cols = len(all_iids)

        # self.uid_encoder = LabelEncoder().fit(all_uids)
        # self.iid_encoder = LabelEncoder().fit(all_iids)
        self.uid_encoder = PDLabelEncoder().fit(all_uids)
        self.iid_encoder = PDLabelEncoder().fit(all_iids)

    def add_encoded_cols(self, df):
        df = df.assign(
            **{self.uid_col: self.uid_encoder.transform(df[self.uid_source_col].values.astype(str)),
               self.iid_col: self.iid_encoder.transform(df[self.iid_source_col].values.astype(str))})
        return df

    def build_sparse_interaction_matrix(self, df):
        """
        :param df:
        :return: the sparse matrix populated with interactions, of shape (n_users, n_items)
            of the source DF (which which this builder with initialized
        """

        df = self.remove_unseen_labels(df)

        mat = sp.coo_matrix(
            (df[self.rating_source_col].values,
             (self.uid_encoder.transform(df[self.uid_source_col].values),
              self.iid_encoder.transform(df[self.iid_source_col].values))),
            shape=(self.n_rows, self.n_cols),
            dtype=np.float32).tocsr()

        return mat

    def remove_unseen_labels(self, df):
        # new_u = ~df[self.uid_source_col].isin(self.uid_encoder.classes_)
        new_u = self.uid_encoder.find_new_labels(df[self.uid_source_col])
        # new_i = ~df[self.iid_source_col].isin(self.iid_encoder.classes_)
        new_i = self.iid_encoder.find_new_labels(df[self.iid_source_col])
        percent_new_u = np.mean(new_u)
        percent_new_i = np.mean(new_i)
        if percent_new_u > 0.0 or percent_new_i > 0.0:
            logger.info(
                'Discarding %.1f%% samples with unseen '
                'users(%d) / unseen items(%d) from DF(len: %s).' % \
                (100 * np.mean(new_u | new_i), np.sum(new_u), np.sum(new_i), len(df)))
            return df[~new_u & ~new_i].copy()
        else:
            return df

    def predictions_df_to_sparse_ranks(self, preds_df):
        preds_all = self.build_sparse_interaction_matrix(preds_df)
        return self.predictions_to_ranks(preds_all)

    @staticmethod
    def predictions_to_ranks(sp_preds):
        # convert prediction matrix to ranks matrix
        ranks_mat = sp_preds.tocsr().copy()
        ranks_mat.sort_indices()
        for i in range(ranks_mat.shape[0]):
            ranks_mat.data[ranks_mat.indptr[i]: ranks_mat.indptr[i + 1]] = \
                np.argsort(
                    np.argsort(
                        -ranks_mat.data[ranks_mat.indptr[i]: ranks_mat.indptr[i + 1]])). \
                    astype(np.float32)
        return ranks_mat

    @staticmethod
    def _filt_ranks_mat_by_filt_mat(inds, ranks_mat, filter_mat):
        # apparently a faster implementation of this would be by using lil matrix format

        ranks_mat.sort_indices()
        filter_mat.sort_indices()
        filt_ranks = filter_mat.copy()
        filt_ranks.data *= 0  # remove actual data
        filt_ranks.data += int(filt_ranks.shape[1] / 2)  # add number of columns / 2 - meaning chance rank
        for i in inds:
            f_s = filt_ranks.indptr[i]
            f_e = filt_ranks.indptr[i + 1]
            r_s = ranks_mat.indptr[i]
            r_e = ranks_mat.indptr[i + 1]

            f_cols = filt_ranks.indices[f_s: f_e]
            r_cols = ranks_mat.indices[r_s: r_e]

            mask_filt = np.isin(f_cols, r_cols)
            mask_ranks = np.isin(r_cols, f_cols)

            # adding 1 to differentiate from 0 as first rank
            filt_ranks.data[f_s: f_e][mask_filt] = ranks_mat.data[r_s: r_e][mask_ranks] + 1

        filt_ranks.eliminate_zeros()
        filt_ranks.data -= 1
        filt_ranks = filt_ranks.tocoo()
        return filt_ranks

    @staticmethod
    def crop_rows(mat, inds_stay):
        mat = mat.tocoo()
        min_data = np.min(mat.data)
        mat.data += min_data
        mat.data[~np.in1d(mat.row, inds_stay)] *= 0
        mat.eliminate_zeros()
        mat.data -= min_data
        return mat.tocsr()

    @staticmethod
    def crop_rows_continuous(mat, ind_start, ind_end):
        mat = mat.tocsr().copy()
        mat.sort_indices()
        mat.data += 1
        mat.data[:mat.indptr[ind_start]] *= 0
        mat.data[(mat.indptr[ind_end + 1] + 1):] *= 0
        mat.eliminate_zeros()
        mat.data -= 1
        return mat

    @classmethod
    def filter_all_ranks_by_sparse_selection(cls, sparse_filter_mat, all_recos_ranks_mat):
        """
        generates rankings for a an evaluation of a dataset (test set), relative to all valid predictions

        :param sparse_filter_mat: sparse matrix of test observations (ground truth)
        :param all_recos_ranks_mat: sparse matrix of all ranked predictions
        :return: sparse matrix of ranks of the predictions for GT observations in the full prediction matrix
        """
        filter_mat = sparse_filter_mat.tocsr()
        ranks_mat = all_recos_ranks_mat.tocsr()
        filter_mat.sort_indices()
        ranks_mat.sort_indices()

        assert ranks_mat.shape == filter_mat.shape

        n_rows = ranks_mat.shape[0]
        res = []
        with Pool(N_CPUS) as pool:
            for ind_batch in batch_generator(np.arange(n_rows),
                                             n=int(n_rows / N_CPUS)):
                res.append(pool.apply_async(
                    cls._filt_ranks_mat_by_filt_mat,
                    args=(
                        ind_batch,
                        cls.crop_rows_continuous(ranks_mat, ind_batch[0], ind_batch[-1]),
                        cls.crop_rows_continuous(filter_mat, ind_batch[0], ind_batch[-1]))))
            ret = [r.get(timeout=3600) for r in res]

        data = np.concatenate([r.data for r in ret])
        row = np.concatenate([r.row for r in ret])
        col = np.concatenate([r.col for r in ret])
        filt_ranks = sp.coo_matrix((data, (row, col)), shape=ranks_mat.shape)

        # ranks are 0 based, and float32 in LFM
        filt_ranks = filt_ranks.astype(np.float32).tocsr()

        return filt_ranks



import warnings
from abc import abstractmethod, ABC
from functools import partial
from itertools import repeat
from multiprocessing import Process, Queue
from multiprocessing.pool import ThreadPool, Pool
from queue import PriorityQueue
from threading import Thread

import numpy as np
import pandas as pd
import scipy.stats

from ml_recsys_tools.recommenders.recommender_base import BaseDFSparseRecommender
from ml_recsys_tools.utils.parallelism import N_CPUS


RANK_COMBINATION_FUNCS = {
    'mean': np.mean,
    'max': np.max,
    'min': np.min,
    'gmean': scipy.stats.gmean,
    'hmean': scipy.stats.hmean
}


def calc_dfs_and_combine_scores(calc_funcs, groupby_col, item_col, scores_col,
                                fill_val, combine_func='hmean', n_threads=1,
                                parallelism='process'):
    """
    combine multiple dataframes by voting on prediction rank

    :param calc_funcs: functions that return the dataframes to be combined
    :param combine_func: defaults 'hmean', the functions that is used to combine the predictions
        (can be callable line np.mean or a string that is assumed to be
        a key in rank_combination_functions mapping
    :param fill_val: rank to be assigned to NaN prediction values
        (items appearing in some dataframes but not in others)
    :param groupby_col: the column of the entities for which the ranking is calculated (e.g. users)
    :param item_col: the column of the entities to be ranked (items)
    :param scores_col: the column of the scores to be ranked (predictions)
    :param n_threads: number of calculation threads
    :param parallelism: type of parallelism (processes or threads)
    :return: a combined dataframe of the same format as the dataframes created by the calc_funcs
    """
    # set up
    multiproc = 'process' in parallelism
    _END = 'END'
    q_in = Queue()
    q_out = Queue() if multiproc else PriorityQueue()
    rank_cols = ['rank_' + str(i) for i in range(len(calc_funcs))]
    n_jobs = len(calc_funcs)
    n_workers = min(n_threads, n_jobs)
    if not callable(combine_func):
        combine_func = RANK_COMBINATION_FUNCS[combine_func]

    jitter = lambda: np.random.rand()

    def _calc_df_and_add_rank_score(i):
        df = calc_funcs[i]()
        df = df.drop_duplicates()

        # another pandas bug workaround
        df[groupby_col] = df[groupby_col].astype(str, copy=False)
        df[item_col] = df[item_col].astype(str, copy=False)
        df[scores_col] = df[scores_col].astype(float, copy=False)

        df[rank_cols[i]] = df. \
            reset_index(). \
            groupby(groupby_col)[scores_col].\
            rank(ascending=False)  # resetting index due to pandas bug

        df = df.drop(scores_col, axis=1).set_index([groupby_col, item_col])

        q_out.put((len(df) + jitter(), df))

    def _joiner():
        while True:
            _, df1 = q_out.get()
            if isinstance(df1, str) and df1 == _END:
                break
            _, df2 = q_out.get()
            if isinstance(df2, str) and df2 == _END:
                q_out.put((len(df1) + jitter(), df1))  # put it back
                break
            df_joined = df2.join(df1, how='outer')
            q_out.put((len(df_joined) + jitter(), df_joined))

    def _worker():
        i = q_in.get()
        while i != _END:
            _calc_df_and_add_rank_score(i)
            i = q_in.get()

    if multiproc:
        workers = [Process(target=_worker) for _ in range(n_workers)]
    else:
        workers = [Thread(target=_worker) for _ in range(n_workers)]

    joiner = Thread(target=_joiner)

    # submit and start jobs
    [q_in.put(i) for i in range(n_jobs)] + [q_in.put(_END) for _ in range(n_workers)]
    [j.start() for j in workers + [joiner]]
    [j.join() for j in workers]

    # stop joiner after workers are done by putting END token
    q_out.put((0, _END))
    joiner.join()

    # final reduce (faster to join in couples rather one by one)
    while q_out.qsize() > 1:
        _, df1 = q_out.get()
        _, df2 = q_out.get()
        df_joined = df2.join(df1, how='outer')
        q_out.put((len(df_joined), df_joined))
    # get final result
    _, merged_df = q_out.get()
    merged_df.fillna(fill_val, inplace=True)
    # combine ranks
    merged_df[scores_col] = combine_func(1 / merged_df[rank_cols].values, axis=1)

    # drop temp cols
    merged_df.drop(rank_cols, axis=1, inplace=True)

    return merged_df.reset_index()


class SubdivisionEnsembleBase(BaseDFSparseRecommender, ABC):

    def __init__(self,
                 n_models=1,
                 concurrence_ratio=0.3,
                 concurrency_backend='threads',
                 combination_mode='hmean',
                 na_rank_fill=200,
                 **kwargs):
        self.n_models = n_models
        self.concurrence_ratio = concurrence_ratio
        self.concurrency_backend = concurrency_backend
        self.combination_mode = combination_mode
        self.na_rank_fill = na_rank_fill
        super().__init__(**kwargs)
        self.sub_class_type = None
        self._init_sub_models()

    def get_workers_pool(self, concurrency_backend=None):
        if concurrency_backend is None:
            concurrency_backend = self.concurrency_backend
        if 'thread' in concurrency_backend:
            return ThreadPool(self.n_concurrent())
        elif 'proc' in concurrency_backend:
            return Pool(self.n_concurrent(), maxtasksperchild=3)

    def _init_sub_models(self, **params):
        self.sub_models = [self.sub_class_type(**params.copy())
                           for _ in range(self.n_models)]

    def n_concurrent(self):
        return int(min(np.ceil(self.n_models * self.concurrence_ratio), N_CPUS))

    def _broadcast(self, var):
        if isinstance(var, list) and len(var) == self.n_models:
            return var
        else:
            return [var] * self.n_models

    def set_params(self, **params):
        params = self._pop_set_params(
            params, ['n_models', 'combination_mode',
                     'na_rank_fill', 'concurrence_ratio'])
        # set on self
        super().set_params(**params.copy())
        # init sub models to make sure they're the right object already
        self._init_sub_models()
        # set for each sub_model
        for model in self.sub_models:
            model.set_params(**params.copy())

    @abstractmethod
    def _generate_sub_model_train_data(self, train_obs):
        pass

    @abstractmethod
    def _fit_sub_model(self, args):
        pass

    def fit(self, train_obs, **fit_params):
        self._set_data(train_obs)

        sub_model_train_data_generator = self._generate_sub_model_train_data(train_obs)

        with self.get_workers_pool() as pool:
            self.sub_models = list(
                pool.imap(self._fit_sub_model,
                          zip(range(self.n_models),
                              sub_model_train_data_generator,
                              repeat(fit_params, self.n_models))))
        return self

    def _get_recommendations_flat(self, user_ids, item_ids, n_rec=100, **kwargs):

        calc_funcs = [
            partial(
                self.sub_models[i_model].get_recommendations,
                user_ids=user_ids, item_ids=item_ids,
                n_rec=n_rec, results_format='flat', **kwargs)
            for i_model in range(len(self.sub_models))]

        recos_flat = calc_dfs_and_combine_scores(
            calc_funcs=calc_funcs,
            combine_func=self.combination_mode,
            fill_val=self.na_rank_fill,
            groupby_col=self._user_col,
            item_col=self._item_col,
            scores_col=self._prediction_col,
            n_threads=self.n_concurrent()
        )

        return recos_flat

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          remove_self=True, embeddings_mode=None,
                          simil_mode='cosine', results_format='lists', **kwargs):

        calc_funcs = [
            partial(
                self.sub_models[i_model].get_similar_items,
                item_ids=item_ids, target_item_ids=target_item_ids,
                n_simil=n_simil, remove_self=remove_self,
                embeddings_mode=embeddings_mode, simil_mode=simil_mode,
                results_format='flat')
            for i_model in range(len(self.sub_models))]

        simil_all = calc_dfs_and_combine_scores(
            calc_funcs=calc_funcs,
            combine_func=self.combination_mode,
            fill_val=self.na_rank_fill,
            groupby_col=self._item_col_simil,
            item_col=self._item_col,
            scores_col=self._prediction_col,
            n_threads=self.n_concurrent()
        )

        return simil_all if results_format == 'flat' \
            else self._simil_flat_to_lists(simil_all, n_cutoff=n_simil)

    def predict_for_user(self, user_id, item_ids, rank_training_last=True, sort=True):

        calc_funcs = [
            partial(
                self.sub_models[i_model].predict_for_user,
                user_id=user_id,
                item_ids=item_ids,
                rank_training_last=rank_training_last)
            for i_model in range(len(self.sub_models))]

        df = calc_dfs_and_combine_scores(
            calc_funcs=calc_funcs,
            combine_func=self.combination_mode,
            fill_val=len(item_ids),
            groupby_col=self._user_col,
            item_col=self._item_col,
            scores_col=self._prediction_col,
            n_threads=N_CPUS,
            parallelism='thread'
        )

        if sort:
            df.sort_values(self._prediction_col, ascending=False, inplace=True)

        return df

    # def sub_model_evaluations(self, test_dfs, test_names, include_train=True):
    #     stats = []
    #     reports = []
    #     for m in self.sub_models:
    #         users = m.train_df[self.train_obs.uid_col].unique()
    #         items = m.train_df[self.train_obs.iid_col].unique()
    #         sub_test_dfs = [df[df[self.train_obs.uid_col].isin(users) &
    #                            df[self.train_obs.iid_col].isin(items)] for df in test_dfs]
    #         lfm_report = m.eval_on_test_by_ranking(
    #             include_train=include_train,
    #             test_dfs=sub_test_dfs,
    #             prefix='lfm sub model',
    #             test_names=test_names
    #         )
    #         stats.append('train: %d, test: %s' %
    #                      (len(m.train_df), [len(df) for df in sub_test_dfs]))
    #         reports.append(lfm_report)
    #     return stats, reports


class CombinationEnsembleBase(BaseDFSparseRecommender):

    def __init__(self, recommenders, **kwargs):
        self.recommenders = recommenders
        self.n_recommenders = len(self.recommenders)
        super().__init__(**kwargs)
        self._reuse_data(self.recommenders[0])

    def fit(self, *args, **kwargs):
        warnings.warn('Fit is not supported, recommenders should already be fitted.')




import pandas as pd
import numpy as np

from spotlight.interactions import Interactions
from spotlight.factorization.implicit import ImplicitFactorizationModel
from spotlight.sequence.implicit import ImplicitSequenceModel
from spotlight.sequence.representations import CNNNet

from ml_recsys_tools.recommenders.factorization_base import BaseFactorizationRecommender
from ml_recsys_tools.recommenders.recommender_base import BaseDFSparseRecommender
from ml_recsys_tools.utils.instrumentation import collect_named_init_params
from ml_recsys_tools.utils.similarity import top_N_sorted


def spotlight_interactions_from_sparse(sp_mat):
    sp_mat = sp_mat.tocoo()
    return Interactions(user_ids=sp_mat.row, item_ids=sp_mat.col, ratings=sp_mat.data)


class EmbeddingFactorsRecommender(BaseFactorizationRecommender):

    default_model_params = dict(
        loss='adaptive_hinge',  # 'bpr', 'hinge', 'adaptive hinge'
        embedding_dim=128,
        n_iter=10,
        batch_size=256,
        l2=0.0,
        learning_rate=1e-2,
        num_negative_samples=5)

    default_fit_params = dict(
        verbose=True)

    def _prep_for_fit(self, train_obs, **fit_params):
        # self.toggle_mkl_blas_1_thread(False)
        self._set_data(train_obs)
        self.set_params(**fit_params)
        self.model = ImplicitFactorizationModel(**self.model_params)
        self._set_spotlight_train_data(self.train_mat)

    def _set_spotlight_train_data(self, train_mat):
        self.spotlight_dataset = spotlight_interactions_from_sparse(train_mat)

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.model.fit(self.spotlight_dataset, **self.fit_params)

    def fit_partial(self, train_obs, epochs=1):
        self._set_epochs(epochs)
        if self.model is None:
            self.fit(train_obs)
        else:
            self.model.fit(self.spotlight_dataset)
        return self

    def _set_epochs(self, epochs):
        self.set_params(n_iter=epochs)

    def _predict(self, user_ids, item_ids):
        return self.model.predict(user_ids, item_ids)

    def _get_item_factors(self, mode=None):
        return self.model._net.item_biases.weight.data.numpy().ravel(), \
               self.model._net.item_embeddings.weight.data.numpy()

    def _get_user_factors(self, mode=None):
        return self.model._net.user_biases.weight.data.numpy().ravel(), \
               self.model._net.user_embeddings.weight.data.numpy()

    def _predict_rank(self, test_mat, train_mat=None):
        raise NotImplementedError()


class SequenceEmbeddingRecommender(BaseDFSparseRecommender):

    default_model_params = dict(
        loss='adaptive_hinge',  # 'pointwise', 'bpr', 'hinge', 'adaptive_hinge'
        representation='lstm',  # 'pooling', 'cnn', 'lstm', 'mixture'
        embedding_dim=32,
        n_iter=4,
        batch_size=64,
        l2=0.0,
        learning_rate=2e-3,
        num_negative_samples=25
    )

    default_fit_params = dict(
        max_sequence_length=200,
        timestamp_col='first_timestamp'
    )

    def _interactions_sequence_from_obs(
            self,
            obs,
            timestamp_col='first_timestamp',
            max_sequence_length=10,
            min_sequence_length=None,
            step_size=None,
            **kwargs):

        obs.timestamp_col = timestamp_col

        return Interactions(
            user_ids=self.sparse_mat_builder.uid_encoder.
                transform(obs.user_ids.astype(str)).astype('int32'),
            item_ids=self.sparse_mat_builder.iid_encoder.
                transform(obs.item_ids.astype(str)).astype('int32') + 1,
            ratings=obs.ratings,
            timestamps=obs.timestamps
        ). \
            to_sequence(
            max_sequence_length=max_sequence_length,
            min_sequence_length=min_sequence_length,
            step_size=step_size
        )

    def _prep_for_fit(self, train_obs, **fit_params):
        # self.toggle_mkl_blas_1_thread(False)
        self._set_data(train_obs)
        self._set_fit_params(fit_params)
        self.sequence_interactions = self._interactions_sequence_from_obs(train_obs, **self.fit_params)

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.model = ImplicitSequenceModel(**self.model_params)
        self.model.fit(self.sequence_interactions)

    def _get_recommendations_flat(self, user_ids, n_rec, item_ids=None,
                                  exclude_training=True, **kwargs):

        return self._get_recommendations_exact(
            user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
            exclude_training=exclude_training, results_format='flat')

    def _get_recommendations_exact(self, user_ids, item_ids=None, n_rec=10, exclude_training=True,
                                   results_format='lists'):

        full_pred_mat = self._predict_dense(user_ids, item_ids, exclude_training=exclude_training)

        top_inds, top_scores = top_N_sorted(full_pred_mat, n=n_rec)

        best_ids = item_ids[top_inds] if item_ids is not None else \
            self.sparse_mat_builder.iid_encoder.inverse_transform(top_inds)

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids,
            scores_mat=top_scores, results_format='recommendations_' + results_format)

    def _predict_dense(self, user_ids, item_ids=None, exclude_training=True):

        if item_ids is None:
            item_ids = self.sparse_mat_builder.iid_encoder.classes_

        item_inds = self.sparse_mat_builder.iid_encoder.transform(item_ids)
        user_inds = self.sparse_mat_builder.uid_encoder.transform(user_ids)

        sequences = self.sequence_interactions.sequences

        pred_mat = np.zeros((len(user_inds), len(item_inds)))

        #TODO: very SLOW, try multiproc (batched from caller)

        item_inds_spot = item_inds.reshape(-1, 1) + 1

        for i_row, user_ind in enumerate(user_inds):
            pred_mat[i_row, :] = self.model.predict(sequences[user_ind], item_ids=item_inds_spot)

        if exclude_training:
            exclude_mat_sp_coo = self.train_mat[user_inds, :][:, item_inds].tocoo()
            pred_mat[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

        return pred_mat


class CNNEmbeddingRecommender(SequenceEmbeddingRecommender):

    cnn_default_params = dict(
        kernel_width=9,
        dilation=1,
        num_layers=3,
        nonlinearity='relu',
        residual_connections=True)

    default_fit_params = dict(
        **SequenceEmbeddingRecommender.default_fit_params,
        **cnn_default_params
    )

    def _cnn_net(self, interactions):
        cnn_params = dict(
            num_items=interactions.num_items,
            embedding_dim=self.model_params['embedding_dim'],
            **{k:v for k, v in self.fit_params.items()
               if k in collect_named_init_params(CNNNet)['CNNNet']})

        # updating dilation parameter to be the right tuple
        cnn_params['dilation'] = [cnn_params['dilation']*(2**i)
                                      for i in range(cnn_params['num_layers'])]

        return CNNNet(**cnn_params)

    def _prep_for_fit(self, train_obs, **fit_params):
        super()._prep_for_fit(train_obs, **fit_params)
        self.model_params['representation'] = self._cnn_net(self.sequence_interactions)




from implicit.als import AlternatingLeastSquares
# from implicit.bpr import BayesianPersonalizedRanking
from implicit.nearest_neighbours import bm25_weight

from ml_recsys_tools.recommenders.factorization_base import BaseFactorizationRecommender
from ml_recsys_tools.utils.instrumentation import simple_logger


class ALSRecommender(BaseFactorizationRecommender):

    default_model_params = dict(
        factors=200,
        iterations=2,
        calculate_training_loss=True,
        num_threads=0)

    default_fit_params = dict(
        use_bm25=True,
        bm25_k1=100,
        bm25_b=0.15,
        cg_steps=4,
        regularization=0,
    )

    def _get_item_factors(self, mode=None):
        return None, self.model.item_factors

    def _get_user_factors(self, mode=None):
        return None, self.model.user_factors

    def _prep_for_fit(self, train_obs, **fit_params):
        # self.toggle_mkl_blas_1_thread(True)
        self._set_data(train_obs)
        self.set_params(**fit_params)
        self.model = AlternatingLeastSquares(**self.model_params)
        self.model.cg_steps = self.fit_params['cg_steps']  # not passable to __init__()
        self._set_implib_train_mat(self.train_mat)

    def _set_implib_train_mat(self, train_mat):
        # implib ALS expects matrix in items x users format
        self.implib_train_mat = train_mat.T
        if self.fit_params['use_bm25']:
            self.implib_train_mat = bm25_weight(
                self.implib_train_mat,
                K1=self.fit_params['bm25_k1'],
                B=self.fit_params['bm25_b'])
        self.model.regularization = \
            self.fit_params['regularization'] * self.implib_train_mat.nnz

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.model.fit(self.implib_train_mat)

    def fit_partial(self, train_obs, epochs=1):
        self._set_epochs(epochs)
        if self.model is None:
            self.fit(train_obs)
        else:
            self.model.fit(self.implib_train_mat)
        return self

    def fit_batches(self, train_obs, train_dfs, epochs_per_batch=None, **fit_params):
        self._prep_for_fit(train_obs)
        for i, df in enumerate(train_dfs):
            batch_train_mat = self.sparse_mat_builder.build_sparse_interaction_matrix(df)

            if epochs_per_batch is not None:
                fit_params['iterations'] = epochs_per_batch
            else:
                fit_params['iterations'] = 1
                self.model.cg_steps = 1
            self._set_fit_params(fit_params)

            self._set_implib_train_mat(batch_train_mat)

            simple_logger.info('Fitting batch %d (%d interactions)' % (i, len(df)))
            self.model.fit(self.implib_train_mat)

    def _set_epochs(self, epochs):
        self.set_params(iterations=epochs)
        if self.model is not None:
            self.model.iterations = epochs

    def _predict(self, user_ids, item_ids):
        raise NotImplementedError()

    def _predict_rank(self, test_mat, train_mat=None):
        raise NotImplementedError()





from ml_recsys_tools.data_handlers.interactions_with_features import ObsWithGeoFeatures
from ml_recsys_tools.data_handlers.interaction_handlers_base import RANDOM_STATE
from ml_recsys_tools.recommenders.similarity_recommenders import ItemCoocRecommender
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender
from ml_recsys_tools.recommenders.ensembles_base import SubdivisionEnsembleBase


class GeoGridEnsembleBase(SubdivisionEnsembleBase):

    def __init__(self,
                 geo_grid_params=None,
                 n_lat=3,
                 n_long=3,
                 overlap_margin=0.5,
                 min_interactions=0,
                 **kwargs):
        self.geo_box = geo_grid_params
        self.n_lat = n_lat
        self.n_long = n_long
        self.min_interactions = min_interactions
        self.overlap_margin = overlap_margin
        kwargs['n_models'] = self.n_lat * self.n_long
        super().__init__(**kwargs)

    def set_params(self, **params):
        """
        this is for skopt / sklearn compatibility
        """
        params = self._pop_set_params(
            params, ['n_lat', 'n_long', 'overlap_margin', 'min_interactions'])

        self.n_models = self.n_lat * self.n_long

        super().set_params(**params.copy())

    def _generate_sub_model_train_data(self, train_obs: ObsWithGeoFeatures):

        if self.geo_box is None:
            self.geo_box = {
                'max_lat': train_obs.df_items[train_obs.lat_col].max(),
                'min_lat': train_obs.df_items[train_obs.lat_col].min(),
                'max_long': train_obs.df_items[train_obs.long_col].max(),
                'min_long': train_obs.df_items[train_obs.long_col].min(),
            }

        self.geo_filters = train_obs.calcluate_equidense_geo_grid(
            n_lat=self.n_lat, n_long=self.n_long,
            overlap_margin=self.overlap_margin, geo_box=self.geo_box)

        for geo_filt in self.geo_filters:
            yield train_obs. \
                filter_by_location_rectangle(*geo_filt). \
                sample_observations(
                min_user_hist=self.min_interactions,
                min_item_hist=self.min_interactions,
                random_state=RANDOM_STATE)


class LFMEnsembleBase(LightFMRecommender, SubdivisionEnsembleBase):

    def __init__(self,
                 use_item_features=False,
                 item_features_params=None,
                 **kwargs):
        self.use_item_features = use_item_features
        self.item_features_params = item_features_params
        super().__init__(**kwargs)

    def _init_sub_models(self):
        self.sub_class_type = LightFMRecommender
        super()._init_sub_models(**{'use_sample_weight': self.use_sample_weight,
                                    # 'sparse_mat_builder': self.sparse_mat_builder,
                                    'model_params': self.model_params,
                                    'fit_params': self.fit_params
                                    })
        # self._set_sub_model_params()

    def set_params(self, **params):
        """
        this is for skopt / sklearn compatibility
        """
        params = self._pop_set_params(
            params, ['use_item_features'])

        super().set_params(**params.copy())

    def _fit_sub_model(self, args):
        i_m, obs, fit_params = args

        # external features
        if self.use_item_features:
            fit_params['external_features'] = \
                obs.get_item_features_for_obs(**self.item_features_params)

        # self.sub_models[i_m] = sub_model_fit_func(self.sub_models[i_m], obs)
        self.sub_models[i_m].fit(obs, **fit_params)
        return self.sub_models[i_m]

    fit = SubdivisionEnsembleBase.fit
    get_similar_items = SubdivisionEnsembleBase.get_similar_items
    _get_recommendations_flat = SubdivisionEnsembleBase._get_recommendations_flat
    predict_for_user = SubdivisionEnsembleBase.predict_for_user


class LFMGeoGridEnsemble(GeoGridEnsembleBase, LFMEnsembleBase):
    pass


class GeoClusteringEnsembleBase(SubdivisionEnsembleBase):

    def _generate_sub_model_train_data(self, train_obs: ObsWithGeoFeatures):
        train_obs.geo_cluster_items(n_clusters=self.n_models)

        labels = train_obs.df_items[train_obs.cluster_label_col].unique()

        for label in labels:
            yield train_obs. \
                filter_by_cluster_label(label)


class LFMGeoClusteringEnsemble(GeoClusteringEnsembleBase, LFMEnsembleBase):
    pass


class CoocEnsembleBase(SubdivisionEnsembleBase, ItemCoocRecommender):

    def _init_sub_models(self, params=None):
        self.sub_class_type = ItemCoocRecommender
        super()._init_sub_models(fit_params=self.fit_params)

    def _fit_sub_model(self, args):
        i_m, obs, fit_params = args

        # self.sub_models[i_m] = sub_model_fit_func(self.sub_models[i_m], obs)
        self.sub_models[i_m].fit(obs, **fit_params)
        return self.sub_models[i_m]


class CoocGeoGridEnsemble(CoocEnsembleBase, GeoGridEnsembleBase):
    pass



from collections import defaultdict

import numpy as np
import pandas as pd
import scipy.sparse as sp
import sklearn.preprocessing
from sklearn.utils import column_or_1d
from sklearn.utils.validation import check_is_fitted
from pandas.core.algorithms import _get_data_algo, _hashtables, _ensure_object
import importlib

pd_cat_module = importlib.import_module(pd.Categorical.__module__)


class PDLabelEncoder(sklearn.preprocessing.LabelEncoder):
    """
    adapted and sped up further from here: https://github.com/scikit-learn/scikit-learn/issues/7432
    a blazingly fast version of encoder that's using Pandas encoding
    which is using hash tables instead of sorted arrays
    """

    def fit(self, y):
        y = column_or_1d(y, warn=True)
        _, self.classes_ = pd.factorize(y, sort=True)
        self._cat_dtype = pd_cat_module.CategoricalDtype(self.classes_)
        self._dtype = self._cat_dtype.categories.dtype
        self._table = self._get_table_for_categories(self.classes_, self._cat_dtype.categories)
        return self

    @staticmethod
    def _get_table_for_categories(values, categories):
        # ripped out of _get_codes_for_values() in pandas Categorical module
        if not pd_cat_module.is_dtype_equal(values.dtype, categories.dtype):
            values = _ensure_object(values)
            categories = _ensure_object(categories)

        (hash_klass, vec_klass), vals = _get_data_algo(values, _hashtables)
        (_, _), cats = _get_data_algo(categories, _hashtables)
        t = hash_klass(len(cats))
        t.map_locations(cats)
        return t

    def transform(self, y, check_labels=True):
        check_is_fitted(self, ['classes_', '_cat_dtype', '_table', '_dtype'])
        y = column_or_1d(y, warn=True)

        ## slower because it creates the hash table every time
        # trans_y = pd.Categorical(y, dtype=self._cat_dtype).codes.copy()
        trans_y = pd_cat_module.coerce_indexer_dtype(
            indexer=self._table.lookup(y.astype(self._dtype)),
            categories=self._cat_dtype.categories)

        if check_labels:
            mask_new_labels = self._new_labels_locs(trans_y)
            if np.any(mask_new_labels):
                raise ValueError("y contains new labels: %s" %
                                 str(np.unique(y[mask_new_labels])))

        return trans_y

    def _new_labels_locs(self, trans_y):
        return trans_y == -1

    def find_new_labels(self, y):
        return self._new_labels_locs(self.transform(y, check_labels=False))

    def __getstate__(self):
        # custom pickling behaviour because the table object is unpickleable
        state = super().__getstate__()
        state['_table'] = None
        return state

    def __setstate__(self, state):
        super().__setstate__(state)
        self._table = self._get_table_for_categories(self.classes_, self._cat_dtype.categories)


class DictLabelEncoder(sklearn.preprocessing.LabelEncoder):
    """
    python dict based version of the hash table based encoder
    somewhat slower but no pandas private parts and no pickling issues
    """

    def fit(self, y):
        self.classes_ = np.array(list(set(y)))
        self._table = defaultdict(
            lambda: -1, {k: i for i, k in enumerate(self.classes_)})
        return self

    def transform(self, y, check_labels=True):
        y = np.array(y)
        trans_y = np.array([self._table[v] for v in y])

        if check_labels:
            mask_new_labels = self._new_labels_locs(trans_y)
            if np.any(mask_new_labels):
                raise ValueError("y contains new labels: %s" %
                                 str(np.unique(y[mask_new_labels])))
        return trans_y

    def _new_labels_locs(self, trans_y):
        return trans_y == -1

    def find_new_labels(self, y):
        return self._new_labels_locs(self.transform(y, check_labels=False))


class NumericBinningEncoder(sklearn.preprocessing.LabelEncoder):
    """
    class for label-encoding a continuous variable by binning
    """
    def __init__(self, n_bins=50):
        super().__init__()
        self.n_bins = n_bins
        self.bins = None

    def fit(self, y):
        percentiles = list(np.linspace(0, 100, num=(self.n_bins + 1)))
        self.bins = np.percentile(y, percentiles[1:])

        if len(np.unique(self.bins)) != len(self.bins):
            self.bins = list(np.linspace(
                np.min(y) - 0.001, np.max(y) + 0.001, num=(self.n_bins + 1)))

        return self

    def transform(self, y):
        inc_bins = list(self.bins)
        inc_bins[0] = min(inc_bins[0], np.min(y))
        inc_bins[-1] = max(inc_bins[-1], np.max(y))

        y_binned = pd.cut(y, bins=inc_bins, labels=False, include_lowest=True)
        y_ind = y_binned.astype(int, copy=False)
        return y_ind


class NumericBinningBinarizer(sklearn.preprocessing.LabelBinarizer):
    """
    class for one-hot encoding a continuous variable by binning
    """
    def __init__(self, n_bins=50, spillage=2, **kwargs):
        """
        :param n_bins: number of bins
        :param spillage:
            number of neighbouring bins that are also activated in
            order to preserve some "proximity" relationship, default to 2
            e.g. for spillage=1 a result vec would be [.. 0, 0, 0.25, 0.5, 1, 0.5, 0.25, 0, 0 ..]
        :param kwargs:
        """
        super().__init__(**kwargs)
        self._spillage = spillage
        self._binner = NumericBinningEncoder(n_bins=n_bins)

    def fit(self, y):
        self._binner.fit(y)
        super().fit(range(len(self._binner.bins)))
        return self

    def transform(self, y):
        y_binned = self._binner.transform(y)
        binarized = super().transform(y_binned)
        if self._spillage:
            for i in range(1, self._spillage + 1):
                binarized += super().transform(y_binned + i) / 2**i
                binarized += super().transform(y_binned - i) / 2**i
        return binarized


class GrayCodesNumericBinarizer(sklearn.preprocessing.LabelBinarizer):
    """
    class for encoding a continuous variable by binning and binarizing to Gray codes.
    Using Gray codes might result in better preservation of proximity (in terms
    of euclidean or cosine distance). In addition, in the resulting representations
    the different columns are also acting as a sort of a binary FFT / DCT
    which may surface valuable information for learning.

    and example of Gray codes for 8 bins:
        array([[0., 0., 0.],
               [1., 0., 0.],
               [1., 1., 0.],
               [0., 1., 0.],
               [0., 1., 1.],
               [1., 1., 1.],
               [1., 0., 1.],
               [0., 0., 1.]])
    """
    def __init__(self, n_bins=64, nan_policy='min', **kwargs):
        """
        :param n_bins: number of bins
        :param nan_policy: what values to replace nans with, can be 'min' or a number
        """
        super().__init__(**kwargs)
        self._nan_policy = nan_policy
        self._n_bins = n_bins

    @staticmethod
    def gray_codes_matrix(n):
        n = int(n)
        codes = np.zeros((1 << n, n))
        for i in range(0, 1 << n):
            gray = i ^ (i >> 1)
            while gray:
                codes[i, int(np.log2(gray & ~(gray - 1)))] = 1
                gray &= gray - 1
        return codes

    def fit(self, y):
        self._n_bins = min(self._n_bins, len(np.unique(y)))
        self._binner = NumericBinningEncoder(n_bins=self._n_bins)
        self._codes = self.gray_codes_matrix(np.ceil(np.log2(self._n_bins)))  # next power of two
        if self.sparse_output:
            self._codes = sp.csr_matrix(self._codes)
        self._binner.fit(self._replace_nans(y))
        super().fit(range(len(self._binner.bins)))
        return self

    def _replace_nans(self, y):
        y_no_nans = y.copy()
        if isinstance(self._nan_policy, (np.int_, np.float_, int, float)):
            pass
        elif self._nan_policy=='min':
            self._nan_policy = np.nanmin(y)
        else:
            raise NotImplementedError()
        y_no_nans[np.isnan(y)] = self._nan_policy
        return y_no_nans

    def transform(self, y):
        return self._codes[self._binner.transform(self._replace_nans(y))]

    def inverse_transform(self, Y, **kwargs):
        raise NotImplementedError()


"""
This is an example on datasets-1M demonstrating some additional recommenders
"""

from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
import pandas as pd
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()
ratings_df = pd.read_csv(rating_csv_path)

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid')
train_obs, test_obs = obs.split_train_test(ratio=0.2)

# train and evaluate a Cooccurrence recommender (fast and deterministic)
from ml_recsys_tools.recommenders.similarity_recommenders import ItemCoocRecommender

item_cooc_rec = ItemCoocRecommender()
item_cooc_rec.fit(train_obs)
print(item_cooc_rec.eval_on_test_by_ranking(test_obs, prefix='item cooccurrence '))

# train and evaluate an ALS recommender (fast)
from ml_recsys_tools.recommenders.implib_recommenders import ALSRecommender

als_rec = ALSRecommender()
als_rec.fit(train_obs)
print(als_rec .eval_on_test_by_ranking(test_obs, prefix='als '))




from collections import OrderedDict

import numpy as np
import pandas as pd
from lightfm.evaluation import precision_at_k, recall_at_k, auc_score, reciprocal_rank

from ml_recsys_tools.utils.instrumentation import LogCallsTimeAndOutput
from ml_recsys_tools.utils.parallelism import N_CPUS


class ModelMockRanksCacher:
    """
    this is used in order to use lightfm functions without copying them out and rewriting
    this makes possible to:
        - not calculate ranks every time from scratch (if you have them precalculated)
        - use the functions to score non lightfm models
    """

    def __init__(self, cached_mat):
        self.cached_mat = cached_mat

    def predict_rank(self, *args, **kwargs):
        return self.cached_mat


def mean_scores_report(model, datasets, dataset_names, k=10):
    ranks_list = [model.predict_rank(dataset, num_threads=N_CPUS) for dataset in datasets]
    return mean_scores_report_on_ranks(ranks_list, datasets, dataset_names, k)


def mean_scores_report_on_ranks(ranks_list, datasets, dataset_names, k=10):
    data = []
    for ranks, dataset in zip(ranks_list, datasets):
        full_report = RanksScorer(ranks_mat=ranks, test_mat=dataset, k=k).scores_report()
        res = full_report.describe().loc['mean']
        data.append(res)
    return pd.DataFrame(data=data, index=dataset_names)


class RanksScorer(LogCallsTimeAndOutput):

    def __init__(self, ranks_mat, test_mat, train_mat=None, k=10, verbose=True):
        super().__init__(verbose=verbose)
        self.ranks_mat = ranks_mat
        self.test_mat = test_mat
        self.train_mat = train_mat
        self.k = k
        self.best_ranks_mat = self._best_ranks()

    def _best_ranks(self):
        return best_possible_ranks(self.test_mat)

    def _ranks_kwargs(self):
        return {
            'ranks': self.ranks_mat,
            'test_interactions': self.test_mat,
            'train_interactions': self.train_mat,
            }

    def _best_ranks_kwargs(self):
        return {
            'ranks': self.best_ranks_mat,
            'test_interactions': self.test_mat,
            'train_interactions': self.train_mat,
            }

    def scores_report(self):
        metrics = OrderedDict([
            ('AUC', self.auc()),
            ('reciprocal', self.reciprocal()),
            ('n-MRR@%d' % self.k, self.n_mrr_at_k()),
            ('n-MRR', self.n_mrr()),
            ('n-DCG@%d' % self.k, self.n_dcg_at_k()),
            ('n-precision@%d' % self.k, self.n_precision_at_k()),
            ('precision@%d' % self.k, self.precision_at_k()),
            ('recall@%d' % self.k, self.recall_at_k()),
            ('n-recall@%d' % self.k, self.n_recall_at_k()),
            ('n-i-gini@%d' % self.k, self.n_i_gini_at_k()),
            ('n-diversity@%d' % self.k, self.n_diversity_at_k()),
        ])

        return pd.DataFrame(metrics)[list(metrics.keys())]

    def auc(self):
        return auc_score_on_ranks(**self._ranks_kwargs())

    def reciprocal(self):
        return reciprocal_rank_on_ranks(**self._ranks_kwargs())

    def n_mrr_at_k(self):
        return mrr_norm_on_ranks(**self._ranks_kwargs(), k=self.k)

    def n_mrr(self):
        return mrr_norm_on_ranks(**self._ranks_kwargs())

    def n_dcg_at_k(self):
        return dcg_binary_at_k(**self._ranks_kwargs(), k=self.k) / \
               dcg_binary_at_k(**self._best_ranks_kwargs(), k=self.k)

    def n_precision_at_k(self):
        return precision_at_k_on_ranks(**self._ranks_kwargs(), k=self.k) / \
               precision_at_k_on_ranks(**self._best_ranks_kwargs(), k=self.k)

    def precision_at_k(self):
        return precision_at_k_on_ranks(**self._ranks_kwargs(), k=self.k)

    def n_recall_at_k(self):
        return recall_at_k_on_ranks(**self._ranks_kwargs(), k=self.k) / \
               recall_at_k_on_ranks(**self._best_ranks_kwargs(), k=self.k)

    def recall_at_k(self):
        return recall_at_k_on_ranks(**self._ranks_kwargs(), k=self.k)

    def n_i_gini_at_k(self):
        return (1 - gini_coefficient_at_k(**self._ranks_kwargs(), k=self.k)) / \
               (1 - gini_coefficient_at_k(**self._best_ranks_kwargs(), k=self.k))

    def n_diversity_at_k(self):
        return diversity_at_k(**self._ranks_kwargs(), k=self.k) / \
               diversity_at_k(**self._best_ranks_kwargs(), k=self.k)


def best_possible_ranks(test_mat):
    best_ranks = test_mat.tocsr().copy()
    n_users, n_items = test_mat.shape
    item_inds = np.arange(n_items)
    nnz_counts = best_ranks.getnnz(axis=1)
    best_ranks.data = np.concatenate(
        [np.random.choice(item_inds[:n], n, replace=False) if n else []
         for n in nnz_counts]).astype(np.float32)
    return best_ranks


def chance_ranks(test_mat):
    rand_ranks = test_mat.tocsr().copy()
    n_users, n_items = test_mat.shape
    item_inds = np.arange(n_items)
    nnz_counts = rand_ranks.getnnz(axis=1)
    rand_ranks.data = np.concatenate(
        [np.random.choice(item_inds, n, replace=False)
         for n in nnz_counts]).astype(np.float32)
    return rand_ranks


def precision_at_k_on_ranks(
        ranks, test_interactions, train_interactions=None, k=10, preserve_rows=False):
    return precision_at_k(
        model=ModelMockRanksCacher(ranks.copy()),
        test_interactions=test_interactions,
        train_interactions=train_interactions,
        k=k,
        preserve_rows=preserve_rows)


def recall_at_k_on_ranks(
        ranks, test_interactions, train_interactions=None, k=10, preserve_rows=False):
    return recall_at_k(
        model=ModelMockRanksCacher(ranks.copy()),
        test_interactions=test_interactions,
        train_interactions=train_interactions,
        k=k,
        preserve_rows=preserve_rows,
    )


def auc_score_on_ranks(
        ranks, test_interactions, train_interactions=None, preserve_rows=False):
    return auc_score(
        model=ModelMockRanksCacher(ranks.copy()),
        test_interactions=test_interactions,
        train_interactions=train_interactions,
        preserve_rows=preserve_rows,
    )


def reciprocal_rank_on_ranks(
        ranks, test_interactions, train_interactions=None, preserve_rows=False):
    return reciprocal_rank(
        model=ModelMockRanksCacher(ranks.copy()),
        test_interactions=test_interactions,
        train_interactions=train_interactions,
        preserve_rows=preserve_rows,
    )


def mrr_norm_on_ranks(
        ranks, test_interactions, train_interactions=None, preserve_rows=False, k=None):

    def harmonic_number(n):
        # https://stackoverflow.com/questions/404346/python-program-to-calculate-harmonic-series
        """Returns an approximate value of n-th harmonic number.
           http://en.wikipedia.org/wiki/Harmonic_number
        """
        # Euler-Mascheroni constant
        gamma = 0.57721566490153286060651209008240243104215933593992
        return gamma + np.log(n) + 0.5 / n - 1. / (12 * n ** 2) + 1. / (120 * n ** 4)

    # number of test items in each row + epsilon for subsequent reciprocals / devisions
    total_positives = np.diff(test_interactions.tocsr().indptr)

    reciprocals = ranks.copy()
    reciprocals.data = 1.0 / (ranks.data + 1)

    if k:
        reciprocals.data[ranks.data > (k - 1)] = 0
        total_positives[total_positives > k] = k

    # sum the reciprocals and devide by count of test interactions in each row
    mrr = np.squeeze(np.array(reciprocals.sum(axis=1)))

    # the max mrr is the partial sum of the harmonic series divided by number of items:
    #  1/n * (1/1 + 1/2 + 1/3 ... 1/n) for n = number of items
    max_mrr = harmonic_number(total_positives + 0.001)

    mrr_norm = mrr / max_mrr

    if not preserve_rows:
        mrr_norm = mrr_norm[test_interactions.getnnz(axis=1) > 0]

    return mrr_norm


def dcg_binary_at_k(
        ranks, test_interactions, k=10, train_interactions=None, preserve_rows=False):
    ranks = ranks.copy()

    ranks.data += 1
    ranks.data[ranks.data > k] *= 0
    ranks.eliminate_zeros()

    ranks.data = 1 / (np.log2(ranks.data + 1))

    dcg = np.squeeze(np.array(ranks.sum(axis=1)))

    if not preserve_rows:
        dcg = dcg[test_interactions.getnnz(axis=1) > 0]

    return dcg


def diversity_at_k(ranks, test_interactions, k=10, train_interactions=None, preserve_rows=False):
    """
    Diversity metric:
        calculates the percentage of items that
        were recommended @ k for any user out of all possible items
    """
    ranks = ranks.copy().tocsr()
    ranks.data += 1
    ranks.data[ranks.data > k] *= 0
    ranks.eliminate_zeros()

    cols, counts = np.unique(ranks.indices, return_counts=True)
    n_cols = ranks.shape[1]
    percentage = len(counts) / n_cols

    n_rows = np.sum(test_interactions.getnnz(axis=1) > 0) if not preserve_rows else ranks.shape[0]

    return np.repeat(percentage, n_rows)


def gini_coefficient_at_k(ranks, test_interactions, k=10, train_interactions=None, preserve_rows=False):
    """
    Diversity metric:
        calculates the gini coefficient for the
        counts of recommended items @ k for all users
    """

    def gini(x, w=None):
        # https://stackoverflow.com/questions/48999542/more-efficient-weighted-gini-coefficient-in-python/
        # The rest of the code requires numpy arrays.
        x = np.asarray(x)
        if w is not None:
            w = np.asarray(w)
            sorted_indices = np.argsort(x)
            sorted_x = x[sorted_indices]
            sorted_w = w[sorted_indices]
            # Force float dtype to avoid overflows
            cumw = np.cumsum(sorted_w, dtype=float)
            cumxw = np.cumsum(sorted_x * sorted_w, dtype=float)
            return (np.sum(cumxw[1:] * cumw[:-1] - cumxw[:-1] * cumw[1:]) /
                    (cumxw[-1] * cumw[-1]))
        else:
            sorted_x = np.sort(x)
            n = len(x)
            cumx = np.cumsum(sorted_x, dtype=float)
            # The above formula, with all weights equal to 1 simplifies to:
            return (n + 1 - 2 * np.sum(cumx) / cumx[-1]) / n

    ranks = ranks.copy().tocsr()
    ranks.data += 1
    ranks.data[ranks.data > k] *= 0
    ranks.eliminate_zeros()

    cols, counts = np.unique(ranks.indices, return_counts=True)
    counts = np.concatenate([counts, np.zeros(ranks.shape[1] - len(counts))])

    n_rows = np.sum(test_interactions.getnnz(axis=1) > 0) if not preserve_rows else ranks.shape[0]

    return np.repeat(gini(counts), n_rows)



from lightfm import LightFM
import lightfm.lightfm

from ml_recsys_tools.recommenders.factorization_base import BaseFactorizationRecommender
from ml_recsys_tools.utils.instrumentation import simple_logger
from ml_recsys_tools.utils.parallelism import N_CPUS


# monkey patch print function
def _epoch_logger(s, print_each_n=20):
    try:
        if not int(s.replace('Epoch ', '')) % print_each_n:
            simple_logger.info(s)
    except Exception as e:
        simple_logger.error('Failed in _epoch_logger: %s' % str(e))
        simple_logger.info(s)

lightfm.lightfm.print = _epoch_logger


class LightFMRecommender(BaseFactorizationRecommender):

    default_model_params = {
        'loss': 'warp',
        'learning_schedule': 'adadelta',
        'no_components': 10,
        'max_sampled': 10,
        'item_alpha': 0,
        'user_alpha': 0,
    }

    default_fit_params = {
        'epochs': 100,
        'item_features': None,
        'num_threads': N_CPUS,
        'verbose': True,
    }

    default_external_features_params = dict(add_identity_mat=True)

    def __init__(self,
                 use_sample_weight=False,
                 external_features=None,
                 external_features_params=None, **kwargs):
        self.use_sample_weight = use_sample_weight
        self.external_features = external_features
        self.external_features_params = external_features_params or \
                                        self.default_external_features_params.copy()

        super().__init__(**kwargs)

    def _prep_for_fit(self, train_obs, **fit_params):
        # self.toggle_mkl_blas_1_thread(True)
        # assign all observation data
        self._set_data(train_obs)
        fit_params['sample_weight'] = self.train_mat.tocoo() \
            if self.use_sample_weight else None
        self._set_fit_params(fit_params)
        self._add_external_features()
        # init model and set params
        self.model = LightFM(**self.model_params)

    def _add_external_features(self):
        if self.external_features is not None:
            self.external_features_mat = self.external_features.\
                fit_transform_ids_df_to_mat(
                    items_encoder=self.sparse_mat_builder.iid_encoder,
                    **self.external_features_params)
            simple_logger.info('External item features matrix: %s' %
                            str(self.external_features_mat.shape))

        # add external features if specified
        self.fit_params['item_features'] = self.external_features_mat
        if self.external_features_mat is not None:
            simple_logger.info('Fitting using external features mat: %s'
                               % str(self.external_features_mat.shape))

    def fit(self, train_obs, **fit_params):
        self._prep_for_fit(train_obs, **fit_params)
        self.model.fit(self.train_mat, **self.fit_params)
        return self

    def fit_partial(self, train_obs, epochs=1):
        self._set_epochs(epochs)
        if self.model is None:
            self.fit(train_obs)
        else:
            self.model.fit_partial(self.train_mat)
        return self

    def fit_batches(self, train_obs, train_dfs, epochs_per_batch=None, **fit_params):
        self._prep_for_fit(train_obs)
        for i, df in enumerate(train_dfs):
            batch_train_mat = self.sparse_mat_builder.build_sparse_interaction_matrix(df)

            if epochs_per_batch is not None:
                fit_params['epochs'] = epochs_per_batch

            fit_params['sample_weight'] = batch_train_mat.tocoo() \
                if self.use_sample_weight else None

            self._set_fit_params(fit_params)

            simple_logger.info('Fitting batch %d (%d interactions)' % (i, len(df)))
            self.model.fit_partial(batch_train_mat, **self.fit_params)

    def _set_epochs(self, epochs):
        self.set_params(epochs=epochs)

    def set_params(self, **params):
        params = self._pop_set_params(
            params, ['use_sample_weight', 'external_features', 'external_features_params'])
        super().set_params(**params)

    def _get_item_factors(self, mode=None):

        n_items = len(self.sparse_mat_builder.iid_encoder.classes_)

        biases, representations = self.model.get_item_representations(self.fit_params['item_features'])

        if mode is None:
            pass  # default mode

        elif mode == 'external_features':
            external_features_mat = self.external_features_mat

            assert external_features_mat is not None, \
                'Must define and add a feature matrix for "external_features" similarity.'

            representations = external_features_mat

        elif (mode == 'no_features') and (self.fit_params['item_features'] is not None):

            simple_logger.info('LightFM recommender: get_similar_items: "no_features" mode '
                               'assumes ID mat was added and is the last part of the feature matrix.')

            assert self.model.item_embeddings.shape[0] > n_items, \
                'Either no ID matrix was added, or no features added'

            representations = self.model.item_embeddings[-n_items:, :]

        else:
            raise ValueError('Uknown representation mode: %s' % mode)

        return biases, representations

    def _get_user_factors(self, mode=None):
        return self.model.get_user_representations()

    def _predict(self, user_ids, item_ids):
        return self.model.predict(user_ids, item_ids,
                                  item_features=self.fit_params['item_features'],
                                  num_threads=self.fit_params['num_threads'])


    def _predict_rank(self, test_mat, train_mat=None):
        return self.model.predict_rank(
            test_interactions=test_mat,
            train_interactions=train_mat,
            item_features=self.fit_params['item_features'],
            num_threads=self.fit_params['num_threads'])




import base64
import requests
import time


def singleton(cls):
    instances = {}

    def wrapper(*args, **kwargs):
        if cls not in instances:
            instances[cls] = cls(*args, **kwargs)
        return instances[cls]

    return wrapper


@singleton
class AuthSettings:
    def __init__(self, auth_uri, client_id, secret, scopes):
        self.auth_service_url = auth_uri
        self.client_id = client_id
        self.secret = secret
        self.scopes = scopes


@singleton
class Request:
    """
    """

    def __init__(self, auth_settings, cache):
        self.read_token = ''
        self.auth_service_url = auth_settings.auth_service_url
        self.client_id = auth_settings.client_id
        self.secret = auth_settings.secret
        self.scopes = auth_settings.scopes
        self.cache = cache

    def _load_read_token(self):
        authorization = 'Basic ' + \
                        base64.b64encode(('%s:%s' %
                                          (self.client_id, self.secret)).encode()).decode()
        response = requests.post(self.auth_service_url + '/connect/token',
                                 headers={'Authorization': authorization},
                                 data={'scope': ' '.join(self.scopes),
                                       'grant_type': 'client_credentials'})
        if response.ok:
            oauth_response = response.json()
            access_token = oauth_response['access_token']
            token_type = oauth_response['token_type']
            read_token = token_type + ' ' + access_token
            self._set_cached(self.client_id, read_token, oauth_response['expires_in'] - 60)
        else:
            raise RequestFailure('Failed to retrieve OAuth token.', response=response)
        self.read_token = read_token
        return read_token

    def __call__(self, func):
        def wrapped_func(*args, **kwargs):
            # print(self.read_token)
            self.read_token = self._get_cache(self.client_id)
            if not self.read_token:
                self._load_read_token()
            kwargs = kwargs or {}
            kwargs['headers'] = kwargs.get('headers', {})
            kwargs['headers'].update({'Authorization': self.read_token})
            return func(*args, **kwargs)

        return wrapped_func

    def _get_cache(self, client_id):
        return self.cache.get(client_id)

    def _set_cached(self, client_id, token, expiry):
        return self.cache.set(client_id, token, expiry)


@singleton
class DictCache:
    def __init__(self):
        self.dict = {}

    def not_expired(self, cached_dict):
        return time.time() < cached_dict['expiry']

    def get(self, key, default=None):
        cached = self.dict.get(key, default)
        if cached is not default and self.not_expired(cached):
            return cached['value']
        else:
            return default

    def set(self, key, value, expiry=300):
        self.dict[key] = {'value': value,
                          'expiry': time.time() + expiry}


class RequestFailure(RuntimeError):
    def __init__(self, message, response):
        super(RequestFailure, self).__init__(message)
        self.response = response

# #usage
# CLIENT_ID = 'client-id'
# SECRET = 'secret'
# SCOPES = ['scope1', 'scope1']
# AUTH_SERVICE_URL = 'https://url.com'
#
# settings = AuthSettings(AUTH_SERVICE_URL, CLIENT_ID, SECRET, SCOPES)
# mem_cache = DictCache()
# api_request = Request(settings, mem_cache)
#
# @api_request
# def oauth_get(url, **kwargs):
#     # no 401 403 handler here
#     response = requests.get(url, **kwargs)
#     # print(response)
#     return response
#
# @api_request
# def oauth_post(url, **kwargs):
#     # no 401 403 handler here
#     return requests.post(url, **kwargs)



import numpy as np
import scipy.sparse as sp
import warnings
from functools import partial
from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances
from ml_recsys_tools.utils.parallelism import map_batches_multiproc
from ml_recsys_tools.utils.instrumentation import log_time_and_shape


def _row_ind_mat(ar):
    # returns a matrix of column indexes of the right shape to enable indexing
    return np.indices(ar.shape)[0]


def top_N_unsorted(mat, n):
    # returns top N values and their indexes for each row in a matrix (axis=1)
    # results are unsorted (to save on sort, when only filtering is needed)

    n = np.min([n, mat.shape[-1]])

    top_inds = np.argpartition(mat, -n)[:, -n:]

    top_values = mat[_row_ind_mat(top_inds), top_inds]

    return np.array(top_inds), np.array(top_values)


def _argsort_mask_descending(mat):
    # gets index mask for sorting a matrix by last axis (sorts rows) in descending order
    sort_inds = (_row_ind_mat(mat), np.argsort(-mat, axis=1))
    return sort_inds


def top_N_sorted(mat, n):
    # returns sorted top N elements and indexes in each row of matrix mat

    top_inds, top_values= top_N_unsorted(mat, n)

    sort_inds = _argsort_mask_descending(top_values)

    return top_inds[sort_inds], top_values[sort_inds]


def _top_N_similar(source_inds, source_mat, target_mat, n,
                   exclude_mat_sp=None, source_biases=None, target_biases=None,
                   simil_mode='cosine'):
    """
    for each row in specified inds in source_mat calculates top N similar items in target_mat
    :param source_inds: indices into source mat
    :param source_mat: matrix of features for similarity calculation (left side)
    :param target_mat: matrix of features for similarity calculation (right side)
    :param n: number of top elements to retreive
    :param exclude_mat_sp: a sparse matrix with interactions to exclude
    :param source_biases: bias terms for source_mat
    :param target_biases: bias terms for target_mat
    :param simil_mode: type of similarity calculation:
        'cosine' dot product of normalized matrices (each row sums to 1), without biases
        'dot' regular dot product, without normalization
    :return:
    """

    if simil_mode == 'cosine':
        scores = cosine_similarity(source_mat[source_inds, :], target_mat)

    elif simil_mode == 'euclidean':
        scores = 1 / (euclidean_distances(source_mat[source_inds, :], target_mat) + 0.001)

    elif simil_mode == 'dot':
        scores = np.dot(source_mat[source_inds, :], target_mat.T)

        if source_biases is not None:
            scores = (scores.T + source_biases[source_inds]).T

        if target_biases is not None:
            scores += target_biases

        if sp.issparse(scores):
            scores = scores.toarray()
        else:
            scores = np.array(scores)

    else:
        raise NotImplementedError('unknown similarity mode')

    if exclude_mat_sp is not None:
        exclude_mat_sp_coo = exclude_mat_sp[source_inds, :].tocoo()
        scores[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

    best_inds, best_scores = top_N_unsorted(scores, n)

    sort_inds = _argsort_mask_descending(best_scores)

    return best_inds[sort_inds], best_scores[sort_inds]


def most_similar(source_ids, n, source_encoder, source_mat, source_biases=None,
                 target_ids=None, target_encoder=None, target_mat=None, target_biases=None,
                 exclude_mat_sp=None,
                 chunksize=1000, simil_mode='cosine'):
    """
    multithreaded batched version of _top_N_similar() that works with IDs instead of indices
    for each row in specified IDS in source_mat calculates top N similar items in target_mat

    :param source_ids: IDS of query items in source mat
    :param n: number of top items to find for each query item
    :param remove_self: whether to remove first element - for cases when
        source elements are present in target_mat (self similarity)
    :param source_encoder: encoder for transforming IDS to indeces in source_mat
    :param source_mat: features matrix for query items
    :param source_biases: biases for query items
    :param target_ids: subset of target ids to be considered
    :param target_encoder: encoder for transforming IDS to indeces in target_mat
    :param target_mat: features matrix for target items
    :param target_biases: biases for target items
    :param exclude_mat_sp: a sparse mat with interactions to exclude (e.g. training mat)
    :param chunksize: chunksize for batching (in term of query items)
    :param simil_mode: mode of similarity calculation:
        'cosine' dot product of normalized matrices (each row sums to 1), without biases
        'dot' regular dot product, without normalization, with added biases if supplied
        'euclidean' inverse of euclidean distance
    :return:
        best_ids - matrix (n_ids, N) of N top items from target_mat for each item in IDS of source_mat
        best_scores - similarity scores for best_ids (n_ids, N)
    """

    if target_mat is None:
        target_mat = source_mat
        target_encoder = source_encoder
        target_biases = source_biases

    # to index
    source_inds = source_encoder.transform(np.array(source_ids, dtype=str))

    if target_ids is None:
        target_ids = target_encoder.classes_

    target_inds = target_encoder.transform(np.array(target_ids, dtype=str))
    target_inds.sort()

    chunksize = int(35000 * chunksize / max(source_mat.shape))

    calc_func = partial(
        _top_N_similar,
        source_mat=source_mat,
        target_mat=target_mat[target_inds, :],  # only the relevant submatrix
        exclude_mat_sp=exclude_mat_sp[:, target_inds] if exclude_mat_sp is not None else None,
        n=n,
        source_biases=source_biases,
        target_biases=target_biases[target_inds] if target_biases is not None else None,
        simil_mode=simil_mode)

    ret = map_batches_multiproc(calc_func, source_inds,
                                chunksize=chunksize,
                                threads_per_cpu=2)
    sub_mat_best_inds = np.concatenate([r[0] for r in ret], axis=0)
    best_scores = np.concatenate([r[1] for r in ret], axis=0)

    # back to ids
    best_inds = target_inds[sub_mat_best_inds.astype(int)]
    best_ids = target_encoder.inverse_transform(best_inds.astype(int))

    return best_ids, best_scores


@log_time_and_shape
def custom_row_func_on_sparse(source_ids, source_encoder, target_encoder,
                              sparse_mat, row_func, exclude_mat_sp=None,
                              target_ids=None, chunksize=10000):
    source_inds = source_encoder.transform(np.array(source_ids, dtype=str))

    if target_ids is None:
        target_ids = target_encoder.classes_

    target_inds = target_encoder.transform(np.array(target_ids, dtype=str))
    target_inds.sort()

    sub_mat_ll = sparse_mat. \
                     tocsr()[source_inds, :][:, target_inds].tolil()

    exclude_mat_ll = exclude_mat_sp[source_inds, :][:, target_inds].tolil() \
        if exclude_mat_sp is not None else None

    def top_n_inds_batch(inds_batch):
        nonlocal exclude_mat_ll, sub_mat_ll
        inds_list = []
        vals_list = []
        for i in inds_batch:
            exlclude_inds = np.array(exclude_mat_ll.rows[i]) \
                if exclude_mat_ll is not None else None
            inds_, vals_ = row_func(np.array(sub_mat_ll.rows[i]),
                                    np.array(sub_mat_ll.data[i]),
                                    exlclude_inds)
            inds_list.append(inds_)
            vals_list.append(vals_)
        return np.array(np.stack(inds_list)), np.array(np.stack(vals_list))

    batch_res = map_batches_multiproc(
        top_n_inds_batch, np.arange(sub_mat_ll.shape[0]), chunksize=chunksize)

    sub_mat_best_inds = np.concatenate([r[0] for r in batch_res])
    best_scores = np.concatenate([r[1] for r in batch_res])

    # back to ids
    best_inds = target_inds[sub_mat_best_inds.astype(int)]
    best_ids = target_encoder.inverse_transform(best_inds.astype(int))

    return best_ids, best_scores


@log_time_and_shape
def top_N_sorted_on_sparse(source_ids, target_ids, encoder, sparse_mat,
                           n_top=10, chunksize=10000):
    def _pad_k_zeros(vec, k):
        return np.pad(vec, (0, k), 'constant', constant_values=0)

    def top_n_row(row_indices, row_data, exclude_inds):

        n_min = min(n_top, len(row_data))

        if exclude_inds is not None:
            row_data[row_indices == exclude_inds] = -np.inf

        i_sort = np.argsort(-row_data)[:n_min]

        if n_min == n_top:
            return row_indices[i_sort], row_data[i_sort]
        else:
            return \
                _pad_k_zeros(row_indices[i_sort], n_top - n_min), \
                _pad_k_zeros(row_data[i_sort], n_top - n_min),

    return custom_row_func_on_sparse(
        row_func=top_n_row,
        source_ids=source_ids,
        target_ids=target_ids,
        source_encoder=encoder,
        target_encoder=encoder,
        sparse_mat=sparse_mat,
        chunksize=chunksize,
    )



from functools import partial
from itertools import repeat

from ml_recsys_tools.recommenders.similarity_recommenders import SimilarityDFRecommender
from ml_recsys_tools.utils.parallelism import batch_generator
from ml_recsys_tools.recommenders.ensembles_base import CombinationEnsembleBase, calc_dfs_and_combine_scores


class CombinedRankEnsemble(CombinationEnsembleBase):

    def __init__(self, recommenders, fill_na_val=None,
                 combination_mode='hmean', **kwargs):
        super().__init__(recommenders=recommenders, **kwargs)
        self.fill_na_val = fill_na_val
        self.combination_mode = combination_mode

    def _get_recommendations_flat(self, user_ids, n_rec, item_ids=None,
                                  exclude_training=True, **kwargs):

        calc_funcs = [partial(rec.get_recommendations,
                              user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
                              exclude_training=exclude_training,
                              results_format='flat', **kwargs)
                      for rec in self.recommenders]
        return calc_dfs_and_combine_scores(
            calc_funcs=calc_funcs,
            combine_func=self.combination_mode,
            fill_val=self.fill_na_val if self.fill_na_val else (n_rec + 1),
            groupby_col=self._user_col,
            item_col=self._item_col,
            scores_col=self._prediction_col,
        )

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          n_unfilt=100, results_format='lists', **kwargs):

        calc_funcs = [partial(rec.get_similar_items,
                              item_ids=item_ids, target_item_ids=target_item_ids,
                              n_simil=n_unfilt, results_format='flat', **kwargs)
                      for rec in self.recommenders]

        combined_simil_df = calc_dfs_and_combine_scores(
            calc_funcs=calc_funcs,
            combine_func=self.combination_mode,
            fill_val=self.fill_na_val if self.fill_na_val else (n_unfilt + 1),
            groupby_col=self._item_col_simil,
            item_col=self._item_col,
            scores_col=self._prediction_col)

        return combined_simil_df if results_format == 'flat' \
            else self._simil_flat_to_lists(combined_simil_df, n_cutoff=n_simil)


class CombinedSimilRecoEns(SimilarityDFRecommender):

    def __init__(self,
                 recommenders,
                 similarity_func_params=None,
                 n_unfilt=100,
                 numeric_n_bins=32,
                 combination_mode='hmean',
                 *args, **kwargs):
        self.recommenders = recommenders
        self.similarity_func_params = similarity_func_params
        self.n_unfilt = n_unfilt
        self.combination_mode = combination_mode
        self.numeric_n_bins = numeric_n_bins
        super().__init__(**kwargs)

    def set_params(self, **params):
        params = self._pop_set_params(
            params, ['n_unfilt', 'numeric_n_bins',
                     'combination_mode'])
        super().set_params(**params)

    def _get_similarity_func_params(self):
        if self.similarity_func_params is None:
            return repeat({}, len(self.recommenders))
        elif isinstance(self.similarity_func_params, dict):
            return repeat(self.similarity_func_params, len(self.recommenders))
        elif hasattr(self.similarity_func_params, '__len__'):
            return self.similarity_func_params
        else:
            raise ValueError('Unsupported format for similarity functions parameters: %s'
                             % str(self.similarity_func_params))

    def fit(self, train_obs, batch_size=10000,
            similarity_queue=None, similarity_queue_cutoff=10, **fit_params):

        itemids = self.recommenders[0].all_items

        for i, items in enumerate(batch_generator(itemids, batch_size)):

            calc_funcs = [
                partial(rec.get_similar_items,
                        item_ids=items, n_simil=self.n_unfilt, results_format='flat', **params)
                for rec, params in zip(self.recommenders, self._get_similarity_func_params())]

            simil_df = calc_dfs_and_combine_scores(
                calc_funcs=calc_funcs,
                combine_func=self.combination_mode,
                fill_val=self.n_unfilt + 1,
                groupby_col=self._item_col_simil,
                item_col=self._item_col,
                scores_col=self._prediction_col,
            )

            if similarity_queue:
                similarity_queue.put(
                    self._simil_flat_to_lists(
                        simil_df, n_cutoff=similarity_queue_cutoff))

            if i == 0:
                super().fit(train_obs, simil_df, **fit_params)
            else:
                super().continue_fit(simil_df)

        if similarity_queue:
            similarity_queue.put('END')

        return self


class CascadeEnsemble(CombinationEnsembleBase):

    def __init__(self, recommenders, **kwargs):
        super().__init__(recommenders, **kwargs)
        assert self.n_recommenders == 2, \
            'only 2 recommenders supported'
        assert hasattr(self.recommenders[1], 'predict_on_df'), \
            'no "predict_on_df" for second recommender'

    def _get_recommendations_flat(self, user_ids, n_rec, item_ids=None,
                                  exclude_training=True, **kwargs):
        recos_df = self.recommenders[0].get_recommendations(
            user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
            exclude_training=exclude_training,
            results_format='flat', **kwargs)
        pred_mat_builder = self.get_prediction_mat_builder_adapter(self.sparse_mat_builder)
        return self.recommenders[1].predict_on_df(
            recos_df,
            user_col=pred_mat_builder.uid_source_col,
            item_col=pred_mat_builder.iid_source_col)



"""
This is an example on datasets-1M demonstrating how to combine recommender using ensemble recommenders
"""

from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
import pandas as pd
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()
ratings_df = pd.read_csv(rating_csv_path)

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid')
train_obs, test_obs = obs.split_train_test(ratio=0.2)

# train and test LightFM recommender
from ml_recsys_tools.recommenders.lightfm_recommender import LightFMRecommender

lfm_rec = LightFMRecommender()
lfm_rec.fit(train_obs, epochs=10)

# train and evaluate a Cooccurrence recommender (fast and deterministic)
from ml_recsys_tools.recommenders.similarity_recommenders import ItemCoocRecommender

item_cooc_rec = ItemCoocRecommender()
item_cooc_rec.fit(train_obs)
print(item_cooc_rec.eval_on_test_by_ranking(test_obs, prefix='item cooccurrence '))

# combine LightFM and Cooccurrence recommenders
# using recommendation ranks, and evaluate
from ml_recsys_tools.recommenders.combination_ensembles import CombinedRankEnsemble

comb_ranks_rec = CombinedRankEnsemble(
    recommenders=[lfm_rec, item_cooc_rec])
print(comb_ranks_rec.eval_on_test_by_ranking(test_obs, prefix='combined ranks '))

# combine LightFM and Cooccurrence recommenders
# using similarities, and evaluate
from ml_recsys_tools.recommenders.combination_ensembles import CombinedSimilRecoEns

comb_simil_rec = CombinedSimilRecoEns(
    recommenders=[lfm_rec, item_cooc_rec])
comb_simil_rec.fit(train_obs)
print(comb_simil_rec.eval_on_test_by_ranking(test_obs, prefix='combined simils '))

# combine LightFM, Cooccurrence and Combined Simil recommender
# using recommendation ranks, and evaluate
from ml_recsys_tools.recommenders.combination_ensembles import CombinedRankEnsemble

comb_ranks_simil_rec = CombinedRankEnsemble(
    recommenders=[lfm_rec, item_cooc_rec, comb_simil_rec])
print(comb_ranks_simil_rec.eval_on_test_by_ranking(
    test_obs, prefix='combined ranks and simils '))

# get all recommendations and print a sample
recs_ens = comb_ranks_simil_rec.get_recommendations(
    comb_ranks_simil_rec.all_users, n_rec=5)
print(recs_ens.sample(5))

# get all similarities and print a sample
simils_ens = comb_ranks_simil_rec.get_similar_items(
    comb_ranks_simil_rec.all_items, n_simil=5)
print(simils_ens.sample(10))



import os
from abc import ABC, abstractmethod
from copy import deepcopy
from functools import partial

import numpy as np
import pandas as pd
import pickle

import time

from ml_recsys_tools.utils.logger import simple_logger as logger
from ml_recsys_tools.utils.automl import BayesSearchHoldOut, SearchSpaceGuess
from ml_recsys_tools.evaluation.ranks_scoring import mean_scores_report_on_ranks
from ml_recsys_tools.data_handlers.interaction_handlers_base import InteractionMatrixBuilder, RANDOM_STATE, \
    ObservationsDF
from ml_recsys_tools.utils.instrumentation import LogCallsTimeAndOutput
from ml_recsys_tools.utils.parallelism import map_batches_multiproc
from ml_recsys_tools.utils.similarity import top_N_sorted


class BaseDFRecommender(ABC, LogCallsTimeAndOutput):
    default_model_params = {}
    default_fit_params = {}

    def __init__(self, user_col='userid', item_col='itemid',
                 rating_col='rating', prediction_col='prediction',
                 model_params=None, fit_params=None, verbose=True):
        super().__init__(verbose)
        self._user_col = user_col
        self._item_col = item_col
        self._item_col_simil = item_col + '_source'
        self._rating_col = rating_col
        self._prediction_col = prediction_col
        self.model_params = self.default_model_params.copy()
        self.fit_params = self.default_fit_params.copy()
        self._set_model_params(model_params)
        self._set_fit_params(fit_params)
        # self.train_df = None
        self.model = None

    @classmethod
    def guess_search_space(cls):
        return SearchSpaceGuess(cls).\
            set_from_dict(cls.default_model_params).\
            set_from_dict(cls.default_fit_params)

    @staticmethod
    def toggle_mkl_blas_1_thread(state):
        if state:
            os.environ['MKL_NUM_THREADS'] = '1'
            os.environ['OPENBLAS_NUM_THREADS'] = '1'
        else:
            os.environ.pop('MKL_NUM_THREADS', None)
            os.environ.pop('OPENBLAS_NUM_THREADS', None)

    @staticmethod
    def _dict_update(d, u):
        d = d.copy()
        if u:
            d.update(u)
        return d

    @staticmethod
    def _pop_set_dict(d, params, pop_params):
        for p in pop_params:
            if p in params:
                d[p] = params.pop(p)
        return params

    def _pop_set_params(self, params, pop_params):
        return self._pop_set_dict(self.__dict__, params, pop_params)

    def _set_model_params(self, params):
        self.model_params = self._dict_update(self.model_params, params)

    def _set_fit_params(self, params):
        self.fit_params = self._dict_update(self.fit_params, params)

    def set_params(self, **params):
        """
        this is for skopt / sklearn compatibility
        """
        # pop-set fit_params if provided in bulk
        self._set_fit_params(params.pop('fit_params', {}))
        # pop-set model_params if provided in bulk
        self._set_model_params(params.pop('model_params', {}))
        # pop-set fit_params by keys from default_fit_params if provided flat
        params = self._pop_set_dict(self.fit_params, params, self.default_fit_params.keys())
        # the rest are assumed to be model_params provided flat
        self._set_model_params(params)

    @abstractmethod
    def fit(self, train_obs: ObservationsDF, *args, **kwargs):
        return self

    @abstractmethod
    def get_recommendations(
            self, user_ids=None, item_ids=None, n_rec=10,
            exclude_training=True,
            results_format='lists',
            **kwargs):
        return pd.DataFrame()

    @abstractmethod
    def eval_on_test_by_ranking(
            self, test_dfs, test_names=('',), prefix='rec ',
            include_train=True, items_filter=None, k=10,
            **kwargs):
        return pd.DataFrame()

    def _recos_lists_to_flat(self, recos_lists_df):

        # n_users = len(recos_lists_df)
        n_recos = len(recos_lists_df[self._item_col].iloc[0])

        recos_df_flat = pd.DataFrame({
            self._user_col: recos_lists_df[self._user_col].values.repeat(n_recos),
            self._item_col: np.concatenate(recos_lists_df[self._item_col].values),
            self._prediction_col: np.concatenate(recos_lists_df[self._prediction_col].values),
        })

        return recos_df_flat

    @staticmethod
    def _flat_df_to_lists(df, sort_col, group_col, n_cutoff, target_columns):
        order = [group_col] + target_columns
        return df[order]. \
            sort_values(sort_col, ascending=False). \
            groupby(group_col). \
            aggregate(lambda x: list(x)[:n_cutoff]). \
            reset_index()

    def _recos_flat_to_lists(self, df, n_cutoff=None):
        return self._flat_df_to_lists(
            df,
            n_cutoff=n_cutoff,
            sort_col=self._prediction_col,
            group_col=self._user_col,
            target_columns=[self._item_col, self._prediction_col])

    def _simil_flat_to_lists(self, df, n_cutoff=None):
        return self._flat_df_to_lists(
            df,
            n_cutoff=n_cutoff,
            sort_col=self._prediction_col,
            group_col=self._item_col_simil,
            target_columns=[self._item_col, self._prediction_col])

    def _format_results_df(self, source_vec, target_ids_mat, scores_mat, results_format):
        if 'recommendations' in results_format:
            source_col = self._user_col
            target_col = self._item_col
            scores_col = self._prediction_col
        elif 'similarities' in results_format:
            source_col = self._item_col_simil
            target_col = self._item_col
            scores_col = self._prediction_col
        else:
            raise NotImplementedError('results_format: ' + results_format)

        order = [source_col, target_col, scores_col]

        if 'lists' in results_format:
            return pd.DataFrame({
                source_col: source_vec,
                target_col: list(target_ids_mat),
                scores_col: list(scores_mat)
            })[order]
        elif 'flat' in results_format:
            n_rec = target_ids_mat.shape[1]
            return pd.DataFrame({
                source_col: np.array(source_vec).repeat(n_rec),
                target_col: np.concatenate(list(target_ids_mat)),
                scores_col: np.concatenate(list(scores_mat)),
            })[order]

        else:
            raise NotImplementedError('results_format: ' + results_format)

    def hyper_param_search_on_split_data(
            self, train_data, validation_data, hp_space,
            n_iters=100, random_search=0.9, **kwargs):

        return RecoBayesSearchHoldOut(
                search_space=hp_space,
                pipeline=self,
                loss=None,
                random_ratio=random_search,
                **kwargs).\
            optimize(
                train_data=train_data,
                validation_data=validation_data,
                n_calls=n_iters)

    def pickle_to_file(self, filepath):
        with open(filepath, 'wb') as f:
            pickle.dump(self, f)

    def hyper_param_search(self, train_obs, hp_space, n_iters=100,
                           valid_ratio=0.04, random_search=0.9,
                           valid_split_time_col=None, **kwargs):

        train_obs_bo, valid_obs_bo = train_obs.split_train_test(
            ratio=valid_ratio ** 0.5 if valid_split_time_col is None else valid_ratio,
            users_ratio=valid_ratio ** 0.5 if valid_split_time_col is None else 1,
            time_split_column=valid_split_time_col,
            random_state=RANDOM_STATE)

        return self.hyper_param_search_on_split_data(
            train_data=train_obs_bo, validation_data=valid_obs_bo.df_obs,
            hp_space=hp_space, n_iters=n_iters, random_search=random_search, **kwargs)


class BaseDFSparseRecommender(BaseDFRecommender):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.sparse_mat_builder = None
        self.train_mat = None
        self.external_features_mat = None

    def _set_data(self, train_obs, calc_train_mat=True):
        train_df = train_obs.df_obs
        self.sparse_mat_builder = train_obs.get_sparse_matrix_helper()
        self.all_users = train_df[self.sparse_mat_builder.uid_source_col].unique().astype(str)
        self.all_items = train_df[self.sparse_mat_builder.iid_source_col].unique().astype(str)
        # shuffling because np.unique() returns elements in almost sorted order by counts,
        # and it's probably not a good thing: it changes regional sparsity,
        # and at a later stage might be sampled / iterated in order
        np.random.shuffle(self.all_users)
        np.random.shuffle(self.all_items)
        if calc_train_mat:
            self.train_mat = self.sparse_mat_builder.build_sparse_interaction_matrix(train_df)

    def _reuse_data(self, other):
        self.all_users = other.all_users
        self.all_items = other.all_items
        self.sparse_mat_builder = other.sparse_mat_builder
        self.train_mat = other.train_mat

    def remove_unseen_users(self, user_ids, message_prefix=''):
        return self._filter_array(
            user_ids,
            encoder=self.sparse_mat_builder.uid_encoder,
            message_prefix=message_prefix,
            message_suffix='users that were not in training set.')

    def remove_unseen_items(self, item_ids, message_prefix=''):
        return self._filter_array(
            item_ids,
            encoder=self.sparse_mat_builder.iid_encoder,
            message_prefix=message_prefix,
            message_suffix='items that were not in training set.')

    @staticmethod
    def _filter_array(array, encoder, message_prefix='', message_suffix=''):
        array = np.array(array).astype(str)
        new_labels_mask = encoder.find_new_labels(array)
        n_discard = np.sum(new_labels_mask)
        if n_discard > 0:
            logger.info(
                '%s Discarding %d (out of %d) %s' %
                (message_prefix, int(n_discard), len(array), message_suffix))
        return array[~new_labels_mask]

    # def _remove_training_from_df(self, flat_df):
    #     flat_df = pd.merge(
    #         flat_df, self.train_df,
    #         left_on=[self._user_col, self._item_col],
    #         right_on=[self._user_col, self.sparse_mat_builder.iid_source_col],
    #         how='left')
    #     # keep only data that was present in left (recommendations) but no in right (training)
    #     flat_df = flat_df[flat_df[self._rating_col].isnull()] \
    #         [[self._user_col, self._item_col, self._prediction_col]]
    #     return flat_df

    @staticmethod
    def _remove_self_similarities(flat_df, col1, col2):
        return flat_df[flat_df[col1].values != flat_df[col2].values].copy()

    @staticmethod
    def _eval_on_test_by_ranking_LFM(train_ranks_func, test_tanks_func,
                                     test_dfs, test_names=('',), prefix='',
                                     include_train=True, k=10):
        """
        this is just to avoid the same flow twice (or more)
        :param train_ranks_func: function that return the ranks and sparse mat of training set
        :param test_tanks_func: function that return the ranks and sparse mat of a test set
        :param test_dfs: test dataframes
        :param test_names: test dataframes names
        :param prefix: prefix for this report
        :param include_train: whether to evaluate training or not
        :return: a report dataframe
        """

        # test
        if isinstance(test_dfs, pd.DataFrame):
            test_dfs = [test_dfs]

        report_dfs = []

        # train
        if include_train:
            ranks_mat, sp_mat = train_ranks_func()
            report_dfs.append(mean_scores_report_on_ranks(
                ranks_list=[ranks_mat], datasets=[sp_mat],
                dataset_names=[prefix + 'train'], k=k))

        for test_df, test_name in zip(test_dfs, test_names):
            ranks_mat, sp_mat = test_tanks_func(test_df)
            report_dfs.append(mean_scores_report_on_ranks(
                ranks_list=[ranks_mat], datasets=[sp_mat],
                dataset_names=[prefix + test_name + 'test'], k=k))

        report_df = pd.concat(report_dfs)
        return report_df

    def get_prediction_mat_builder_adapter(self, mat_builder: InteractionMatrixBuilder):
        mat_builder = deepcopy(mat_builder)
        mat_builder.uid_source_col = self._user_col
        mat_builder.iid_source_col = self._item_col
        mat_builder.rating_source_col = self._prediction_col
        return mat_builder

    @abstractmethod
    def _get_recommendations_flat(self, user_ids, n_rec, item_ids=None,
                                  exclude_training=True, **kwargs):
        return pd.DataFrame()

    def get_recommendations(
            self, user_ids=None, item_ids=None, n_rec=10,
            exclude_training=True, results_format='lists',
            **kwargs):

        if user_ids is not None:
            user_ids = self.remove_unseen_users(user_ids, message_prefix='get_recommendations: ')
        else:
            user_ids = self.all_users

        if item_ids is not None:
            item_ids = self.remove_unseen_items(item_ids, message_prefix='get_recommendations: ')

        recos_flat = self._get_recommendations_flat(
            user_ids=user_ids, item_ids=item_ids, n_rec=n_rec,
            exclude_training=exclude_training)

        if results_format == 'flat':
            return recos_flat
        else:
            return self._recos_flat_to_lists(recos_flat, n_cutoff=n_rec)

    def _check_item_ids_args(self, item_ids, target_item_ids):
        item_ids = self.remove_unseen_items(item_ids) \
            if item_ids is not None else self.all_items
        target_item_ids = self.remove_unseen_items(target_item_ids) \
            if target_item_ids is not None else None
        return item_ids, target_item_ids

    def _check_user_ids_args(self, user_ids, target_user_ids):
        user_ids = self.remove_unseen_users(user_ids) \
            if user_ids is not None else self.all_users
        target_user_ids = self.remove_unseen_users(target_user_ids) \
            if target_user_ids is not None else None
        return user_ids, target_user_ids

    def eval_on_test_by_ranking(self, test_dfs, test_names=('',), prefix='rec ',
                                include_train=True, items_filter=None,
                                n_rec=100, k=10):
        @self.logging_decorator
        def relevant_users():
            # get only those users that are present in the evaluation dataframes
            all_test_users = []
            [all_test_users.extend(df[self._user_col].unique().tolist()) for df in test_dfs]
            all_test_users = np.unique(all_test_users)
            return all_test_users

        if isinstance(test_dfs, pd.DataFrame):
            test_dfs = [test_dfs]
        elif isinstance(test_dfs, ObservationsDF):
            test_dfs = [test_dfs.df_obs]

        mat_builder = self.sparse_mat_builder
        pred_mat_builder = self.get_prediction_mat_builder_adapter(mat_builder)

        users = self.remove_unseen_users(relevant_users())

        if include_train:
            recos_flat_train = self.get_recommendations(
                user_ids=users,
                item_ids=None,
                n_rec=min(n_rec, mat_builder.n_cols),
                exclude_training=False,
                results_format='flat')

        recos_flat_test = self.get_recommendations(
            user_ids=users,
            item_ids=items_filter,
            n_rec=min(n_rec, mat_builder.n_cols),
            exclude_training=True,
            results_format='flat')

        ranks_all_test = pred_mat_builder.predictions_df_to_sparse_ranks(
            recos_flat_test)

        @self.logging_decorator
        def _get_training_ranks():
            users_inds = mat_builder.uid_encoder.transform(users)
            users_inds.sort()
            # train_df = self.train_df[self.train_df[self._user_col].isin(users)].copy()
            # sp_train = self.sparse_mat_builder. \
            #     build_sparse_interaction_matrix(train_df).tocsr()
            sp_train = mat_builder.crop_rows(self.train_mat, inds_stay=users_inds)
            sp_train_ranks = pred_mat_builder. \
                filter_all_ranks_by_sparse_selection(
                sp_train, pred_mat_builder.predictions_df_to_sparse_ranks(recos_flat_train))
            return sp_train_ranks, sp_train

        @self.logging_decorator
        def _get_test_ranks(test_df):
            sp_test = self.sparse_mat_builder. \
                build_sparse_interaction_matrix(test_df).tocsr()
            sp_test_ranks = pred_mat_builder. \
                filter_all_ranks_by_sparse_selection(sp_test, ranks_all_test)
            return sp_test_ranks, sp_test

        report_df = self._eval_on_test_by_ranking_LFM(
            train_ranks_func=_get_training_ranks,
            test_tanks_func=_get_test_ranks,
            test_dfs=test_dfs,
            test_names=test_names,
            prefix=prefix,
            include_train=include_train,
            k=k)

        return report_df


class BasePredictorRecommender(BaseDFSparseRecommender):

    @abstractmethod
    def _predict(self, user_ids, item_ids):
        pass

    @abstractmethod
    def _predict_rank(self, test_mat, train_mat=None):
        pass

    def predict_on_df(self, df, exclude_training=True, user_col=None, item_col=None):
        if user_col is not None and user_col != self.sparse_mat_builder.uid_source_col:
            df[self.sparse_mat_builder.uid_source_col] = df[user_col]
        if item_col is not None and item_col != self.sparse_mat_builder.iid_source_col:
            df[self.sparse_mat_builder.iid_source_col] = df[item_col]

        mat_builder = self.sparse_mat_builder
        df = mat_builder.remove_unseen_labels(df)
        df = mat_builder.add_encoded_cols(df)
        df[self._prediction_col] = self._predict(
            df[mat_builder.uid_col].values, df[mat_builder.iid_col].values)

        if exclude_training:
            exclude_mat_sp_coo = \
                self.train_mat[df[mat_builder.uid_col].values, :] \
                    [:, df[mat_builder.iid_col].values].tocoo()
            df[df[mat_builder.uid_col].isin(exclude_mat_sp_coo.row) &
               df[mat_builder.iid_col].isin(exclude_mat_sp_coo.col)][self._prediction_col] = -np.inf

        df.drop([mat_builder.uid_col, mat_builder.iid_col], axis=1, inplace=True)
        return df

    def eval_on_test_by_ranking_exact(self, test_dfs, test_names=('',),
                                      prefix='lfm ', include_train=True, k=10):

        @self.logging_decorator
        def _get_training_ranks():
            ranks_mat = self._predict_rank(self.train_mat)
            return ranks_mat, self.train_mat

        @self.logging_decorator
        def _get_test_ranks(test_df):
            test_sparse = self.sparse_mat_builder.build_sparse_interaction_matrix(test_df)
            ranks_mat = self.model.predict_rank(test_sparse, self.train_mat)
            return ranks_mat, test_sparse

        return self._eval_on_test_by_ranking_LFM(
            train_ranks_func=_get_training_ranks,
            test_tanks_func=_get_test_ranks,
            test_dfs=test_dfs,
            test_names=test_names,
            prefix=prefix,
            include_train=include_train,
            k=k)

    def get_recommendations_exact(
            self, user_ids, item_ids=None, n_rec=10,
            exclude_training=True, chunksize=200, results_format='lists'):

        calc_func = partial(
            self._get_recommendations_exact,
            n_rec=n_rec,
            item_ids=item_ids,
            exclude_training=exclude_training,
            results_format=results_format)

        chunksize = int(35000 * chunksize / self.sparse_mat_builder.n_cols)

        ret = map_batches_multiproc(calc_func, user_ids, chunksize=chunksize)
        return pd.concat(ret, axis=0)

    def _get_recommendations_exact(self, user_ids, item_ids=None, n_rec=10, exclude_training=True,
                                   results_format='lists'):

        full_pred_mat = self._predict_for_users_dense(user_ids, item_ids, exclude_training=exclude_training)

        top_inds, top_scores = top_N_sorted(full_pred_mat, n=n_rec)

        best_ids = self.sparse_mat_builder.iid_encoder.inverse_transform(top_inds)

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids,
            scores_mat=top_scores, results_format='recommendations_' + results_format)

    def _predict_for_users_dense(self, user_ids, item_ids=None, exclude_training=True):
        """
        method for calculating prediction for a grid of users and items.
        this method uses the _predict() method that the subclasses should implement.

        :param user_ids: users ids
        :param item_ids: item ids, when None - all known items are used
        :param exclude_training: when True sets prediction on training examples to -np.inf
        :return: a matrix of predictions (n_users, n_items)
        """

        if item_ids is None:
            item_ids = self.sparse_mat_builder.iid_encoder.classes_

        user_inds = self.sparse_mat_builder.uid_encoder.transform(user_ids)
        item_inds = self.sparse_mat_builder.iid_encoder.transform(item_ids)

        n_users = len(user_inds)
        n_items = len(item_inds)
        user_inds_mat = user_inds.repeat(n_items)
        item_inds_mat = np.tile(item_inds, n_users)

        full_pred_mat = self._predict(user_inds_mat, item_inds_mat).reshape((n_users, n_items))

        if exclude_training:
            exclude_mat_sp_coo = self.train_mat[user_inds, :][:, item_inds].tocoo()
            full_pred_mat[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

        return full_pred_mat


class RecoBayesSearchHoldOut(BayesSearchHoldOut, LogCallsTimeAndOutput):

    def __init__(self, metric='AUC', k=10, **kwargs):
        self.metric = metric
        self.k = k
        super().__init__(**kwargs)

    def _fit_predict_loss(self, pipeline):
        start = time.time()
        pipeline.fit(self.train_data)
        report_df = pipeline.eval_on_test_by_ranking(
            self.validation_data, include_train=False, prefix='', k=self.k)
        loss = 1 - report_df.loc['test', self.metric]
        return loss, self._add_loss_and_time_to_report(
            loss, time.time() - start, report_df=report_df)




import pprint
import numpy as np
import os
import gmaps
import ipywidgets.embed
import colorsys


class ItemsGeoMapper:

    def __init__(self, df_items=None, lat_col='lat', long_col='long'):
        self.df_items = df_items
        self.lat_col = lat_col
        self.long_col = long_col
        self.fig = None
        gmaps.configure(api_key=os.environ['GOOGLE_MAPS_API_KEY'])

    @property
    def loc_cols(self):
        return [self.lat_col, self.long_col]

    def _ceter_location(self):
        return self.df_items[self.loc_cols].apply(np.mean).tolist()

    def _zoom_heuristic(self):

        # https://stackoverflow.com/questions/6048975/google-maps-v3-how-to-calculate-the-zoom-level-for-a-given-bounds
        def gm_heuristic(min_deg, max_deg):
            if max_deg - min_deg > 0:
                return int(np.log(1000 * 360 / (max_deg - min_deg) / 256) / np.log(2))
            else:
                return 14

        ranges = self.df_items[self.loc_cols].apply([np.min, np.max]).as_matrix()

        zoom_level_lat = gm_heuristic(ranges[0][0], ranges[1][0])

        zoom_level_long = gm_heuristic(ranges[0][1], ranges[1][1])

        return min(zoom_level_lat, zoom_level_long)

    def _check_get_view_fig(self):
        if self.fig is None:
            self.fig = gmaps.figure()
            self.fig.widgets.clear()  # clear any history
            self.fig = gmaps.figure(center=self._ceter_location(), zoom_level=self._zoom_heuristic())

    def add_heatmap(self,
                    df_items=None,
                    color=(0, 250, 50),
                    opacity=0.6,
                    sensitivity=5,
                    spread=30, ):

        self._check_get_view_fig()

        if df_items is None:
            df_items = self.df_items

        self.fig.add_layer(
            gmaps.heatmap_layer(
                df_items[self.loc_cols].as_matrix(),
                opacity=opacity,
                max_intensity=sensitivity,
                point_radius=spread,
                dissipating=True,
                gradient=[list(color) + [0],
                          list(color) + [1]]))
        return self

    def add_markers(self,
                    df_items=None,
                    max_markers=1000,
                    color='red',
                    size=2,
                    opacity=1.0,
                    ):

        self._check_get_view_fig()

        if df_items is None:
            df_items = self.df_items

        marker_locs, marker_info = self._markers_with_info(df_items, max_markers=max_markers)

        self.fig.add_layer(gmaps.symbol_layer(
            marker_locs[self.loc_cols].as_matrix(),
            fill_color=color,
            stroke_color=color,
            fill_opacity=opacity,
            stroke_opacity=opacity,
            scale=size,
            info_box_content=marker_info))
        return self

    @staticmethod
    def _markers_with_info(df_items, max_markers):
        marker_locs = df_items.iloc[:max_markers]
        info_box_template = \
            """
            <dl>            
            <dt>{description}</dt>
            </dl>
            """
        marker_info = [
            info_box_template.format(
                description=pprint.pformat(item_data.to_dict()))
            for _, item_data in marker_locs.iterrows()]

        return marker_locs, marker_info

    def draw_listings(self, df_items=None, **kwargs):
        self.add_heatmap(df_items, **kwargs)
        self.add_markers(df_items, **kwargs)
        return self

    def write_html(self, path):
        for w in self.fig.widgets.values():
            if isinstance(w, ipywidgets.Layout) and w.height == '400px':
                w.height = '700px'

        ipywidgets.embed.embed_minimal_html(path, views=[self.fig])
        return self

    @staticmethod
    def random_color():
        return tuple(map(int, np.random.randint(0, 255, 3)))  # because of bug in gmaps type checking

    @staticmethod
    def get_n_spaced_colors(n):
        # max_value = 16581375  # 255**3
        # interval = int(max_value / n)
        # colors = [hex(I)[2:].zfill(6) for I in range(0, max_value, interval)]
        # return [(int(i[:2], 16), int(i[2:4], 16), int(i[4:], 16)) for i in colors]

        HSV_tuples = [(x * 1.0 / n, 1.0, 0.8) for x in range(n)]
        RGB_tuples = list(map(lambda x:
                              tuple(list(map(lambda f: int(f * 255),
                                             colorsys.hsv_to_rgb(*x)))),
                              HSV_tuples))
        return RGB_tuples



"""
This is an example on datasets-1M demonstrating recommenders from spotlight library
"""

from ml_recsys_tools.datasets.prep_movielense_data import get_and_prep_data
import pandas as pd
from ml_recsys_tools.data_handlers.interaction_handlers_base import ObservationsDF

rating_csv_path, users_csv_path, movies_csv_path = get_and_prep_data()
ratings_df = pd.read_csv(rating_csv_path)

obs = ObservationsDF(ratings_df, uid_col='userid', iid_col='itemid', timestamp_col='timestamp')
train_obs, test_obs = obs.split_train_test(ratio=0.2, time_split_column=obs.timestamp_col)


from ml_recsys_tools.recommenders.spotlight_recommenders import EmbeddingFactorsRecommender
emb_rec = EmbeddingFactorsRecommender(model_params=dict(loss='adaptive_hinge', n_iter=1),
          fit_params=dict(max_sequence_length=200, timestamp_col='timestamp'))
# emb_rec.fit(train_obs)
emb_rec.fit_with_early_stop(train_obs, epochs_max=5, epochs_step=1)
print(emb_rec.eval_on_test_by_ranking(test_obs, prefix='implicit embeddings '))


# trying to reproduce this:
# https://github.com/maciejkula/spotlight/tree/master/examples/movielens_sequence
from ml_recsys_tools.recommenders.spotlight_recommenders import SequenceEmbeddingRecommender
seq_rec = SequenceEmbeddingRecommender(
    model_params=dict(n_iter=15, embedding_dim=32, batch_size=32, learning_rate=0.01),
    fit_params=dict(max_sequence_length=200, timestamp_col='timestamp'))
seq_rec.fit(train_obs)
# emb_rec.fit_with_early_stop(train_obs, epochs_max=30, epochs_step=3)
print(seq_rec.eval_on_test_by_ranking(test_obs, prefix='lstm ', include_train=False))

# same experiment: CNNs
from ml_recsys_tools.recommenders.spotlight_recommenders import CNNEmbeddingRecommender
cnn_rec = CNNEmbeddingRecommender(
    fit_params=dict(max_sequence_length=200, timestamp_col='timestamp'))
cnn_rec.fit(train_obs)
print(cnn_rec.eval_on_test_by_ranking(test_obs, prefix='cnn ', include_train=False))



from abc import abstractmethod
from copy import deepcopy

import pandas as pd
import numpy as np
import scipy.sparse as sp

from ml_recsys_tools.data_handlers.interaction_handlers_base import RANDOM_STATE
from ml_recsys_tools.recommenders.recommender_base import BasePredictorRecommender
from ml_recsys_tools.utils.automl import early_stopping_runner
from ml_recsys_tools.utils.logger import simple_logger
from ml_recsys_tools.utils.similarity import most_similar


class BaseFactorizationRecommender(BasePredictorRecommender):

    @abstractmethod
    def _get_item_factors(self, mode=None):
        return  np.array([0]), np.array([0])

    @abstractmethod
    def _get_user_factors(self, mode=None):
        return np.array([0]), np.array([0])

    @abstractmethod
    def _prep_for_fit(self, train_obs, **fit_params):
        pass

    @abstractmethod
    def fit_partial(self, train_obs, **fit_params):
        pass

    @abstractmethod
    def _set_epochs(self, epochs):
        pass

    def _factors_to_dataframe(self, factor_func, include_biases=False):
        b, f = factor_func()
        return pd.DataFrame(
            index=np.arange(f.shape[0]),
            data=np.concatenate([b, f], axis=1) if include_biases else f)

    def user_factors_dataframe(self, include_biases=False):
        return self._factors_to_dataframe(self._get_user_factors, include_biases=include_biases)

    def item_factors_dataframe(self, include_biases=False):
        return self._factors_to_dataframe(self._get_item_factors, include_biases=include_biases)

    def fit_with_early_stop(self, train_obs, valid_ratio=0.04, refit_on_all=False, metric='AUC',
                            epochs_start=0, epochs_max=200, epochs_step=10, stop_patience=10,
                            plot_convergence=True, decline_threshold=0.05, k=10, valid_split_time_col=None):

        # split validation data
        train_obs_internal, valid_obs = train_obs.split_train_test(
            ratio=valid_ratio ** 0.5 if valid_split_time_col is None else valid_ratio,
            users_ratio=valid_ratio ** 0.5 if valid_split_time_col is None else 1,
            time_split_column=valid_split_time_col,
            random_state=RANDOM_STATE)

        self.model = None
        self.model_checkpoint = None
        all_metrics = pd.DataFrame()

        def update_full_metrics_df(cur_epoch, report_df):
            nonlocal all_metrics
            all_metrics = all_metrics.append(
                report_df.rename(index={'test': cur_epoch}))

        def check_point_func():
            if not refit_on_all:
                self.model_checkpoint = deepcopy(self.model)

        def score_func(cur_epoch, step):
            self.fit_partial(train_obs_internal, epochs=step)
            lfm_report = self.eval_on_test_by_ranking(
                valid_obs.df_obs, include_train=False, prefix='', k=k)
            cur_score = lfm_report.loc['test', metric]
            update_full_metrics_df(cur_epoch, lfm_report)
            return cur_score

        best_epoch = early_stopping_runner(
            score_func=score_func,
            check_point_func=check_point_func,
            epochs_start=epochs_start,
            epochs_max=epochs_max,
            epochs_step=epochs_step,
            stop_patience=stop_patience,
            decline_threshold=decline_threshold,
            plot_graph=plot_convergence
        )
        simple_logger.info('Early stop, all_metrics:\n' + str(all_metrics))

        if plot_convergence:
            all_metrics = all_metrics.divide(all_metrics.max())
            all_metrics.plot()
        self.early_stop_metrics_df = all_metrics

        self._set_epochs(epochs=best_epoch)

        if refit_on_all:
            simple_logger.info('Refitting on whole train data for %d epochs' % best_epoch)
            self.fit(train_obs)
        else:
            simple_logger.info('Loading best model from checkpoint at %d epochs' % best_epoch)
            self.model, self.model_checkpoint = self.model_checkpoint, None

        return self

    def get_similar_items(self, item_ids=None, target_item_ids=None, n_simil=10,
                          remove_self=True, embeddings_mode=None,
                          simil_mode='cosine', results_format='lists'):
        """
        uses learned embeddings to get N most similar items

        :param item_ids: vector of item IDs
        :param n_simil: number of most similar items to retrieve
        :param remove_self: whether to remove the the query items from the lists (similarity to self should be maximal)
        :param embeddings_mode: the item representations to use for calculation:
             None (default) - means full representations
             'external_features' - calculation based only external features (assumes those exist)
             'no_features' - calculation based only on internal features (assumed identity mat was part of the features)
        :param simil_mode: mode of similairyt calculation:
            'cosine' (default) - cosine similarity bewtween representations (normalized dot product with no biases)
            'dot' - unnormalized dot product with addition of biases
            'euclidean' - inverse of euclidean distance
            'cooccurance' - no usage of learned features - just cooccurence of items matrix
                (number of 2nd degree connections in user-item graph)
        :param results_format:
            'flat' for dataframe of triplets (source_item, similar_item, similarity)
            'lists' for dataframe of lists (source_item, list of similar items, list of similarity scores)

        :return: a matrix of most similar IDs [n_ids, N], a matrix of score of those similarities [n_ids, N]
        """

        item_ids, target_item_ids = self._check_item_ids_args(item_ids, target_item_ids)

        biases, representations = self._get_item_factors(mode=embeddings_mode)

        best_ids, best_scores = most_similar(
            source_ids=item_ids,
            target_ids=target_item_ids,
            source_encoder=self.sparse_mat_builder.iid_encoder,
            source_mat=representations,
            source_biases=biases,
            n=n_simil+1 if remove_self else n_simil,
            simil_mode=simil_mode,
        )

        simil_df = self._format_results_df(
            item_ids, target_ids_mat=best_ids,
            scores_mat=best_scores, results_format='similarities_flat')

        if remove_self:
            simil_df = self._remove_self_similarities(
                simil_df, col1=self._item_col_simil, col2=self._item_col)

        if 'lists' in results_format:
            simil_df = self._simil_flat_to_lists(simil_df, n_cutoff=n_simil)

        return simil_df

    def get_similar_users(self, user_ids=None, target_user_ids=None, n_simil=10, remove_self=True,
                          simil_mode='cosine'):
        """
        same as get_similar_items but for users
        """
        user_ids, target_user_ids = self._check_user_ids_args(user_ids, target_user_ids)

        user_biases, user_representations = self._get_user_factors()

        best_ids, best_scores = most_similar(
            source_ids=user_ids,
            target_ids=target_user_ids,
            source_encoder=self.sparse_mat_builder.uid_encoder,
            source_mat=user_representations,
            source_biases=user_biases,
            n=n_simil,
            simil_mode=simil_mode,
        )

        simil_df = self._format_results_df(
            user_ids, target_ids_mat=best_ids,
            scores_mat=best_scores, results_format='similarities_flat'). \
            rename({self._item_col_simil: self._user_col}, axis=1)
        # this is UGLY, if this function is ever useful, fix this please (the renaming shortcut)

        if remove_self:
            simil_df = self._remove_self_similarities(
                simil_df, col1=self._user_col, col2=self._item_col)

        simil_df = self._recos_flat_to_lists(simil_df, n_cutoff=n_simil)

        return simil_df

    def _get_recommendations_flat(
            self, user_ids, n_rec, item_ids=None, exclude_training=True,
            item_features_mode=None, use_biases=True):

        user_biases, user_representations = self._get_user_factors()
        item_biases, item_representations = self._get_item_factors(mode=item_features_mode)

        if not use_biases:
            user_biases, item_biases = None, None

        best_ids, best_scores = most_similar(
            source_ids=user_ids,
            target_ids=item_ids,
            source_encoder=self.sparse_mat_builder.uid_encoder,
            target_encoder=self.sparse_mat_builder.iid_encoder,
            source_mat=user_representations,
            target_mat=item_representations,
            source_biases=user_biases,
            target_biases=item_biases,
            exclude_mat_sp=self.train_mat if exclude_training else None,
            n=n_rec,
            simil_mode='dot',
        )

        return self._format_results_df(
            source_vec=user_ids, target_ids_mat=best_ids, scores_mat=best_scores,
            results_format='recommendations_flat')

    def _predict_for_users_dense_direct(self, user_ids, item_ids=None, exclude_training=True):
        """
        method for calculating prediction for a grid of users and items
        directly from the calculated factors. this method is faster for smaller inputs
        an can be further sped up by employing batched multiprocessing (as
        used in similarity / recommendation calculations)

        :param user_ids: users ids
        :param item_ids: item ids, when None - all known items are used
        :param exclude_training: when True sets prediction on training examples to -np.inf
        :return: a matrix of predictions (n_users, n_items)
        """
        if item_ids is None:
            item_ids = self.sparse_mat_builder.iid_encoder.classes_

        user_inds = self.sparse_mat_builder.uid_encoder.transform(user_ids)
        item_inds = self.sparse_mat_builder.iid_encoder.transform(item_ids)

        user_biases, user_factors = self._get_user_factors()
        item_biases, item_factors = self._get_item_factors()

        scores = np.dot(user_factors[user_inds, :], item_factors[item_inds, :].T)

        if user_biases is not None:
            scores = (scores.T + user_biases[user_inds]).T

        if item_biases is not None:
            scores += item_biases[item_inds]

        if sp.issparse(scores):
            scores = scores.toarray()
        else:
            scores = np.array(scores)

        full_pred_mat = scores

        if exclude_training:
            exclude_mat_sp_coo = self.train_mat[user_inds, :][:, item_inds].tocoo()
            full_pred_mat[exclude_mat_sp_coo.row, exclude_mat_sp_coo.col] = -np.inf

        return full_pred_mat

    def predict_for_user(self, user_id, item_ids, rank_training_last=True, sort=True):
        """
        method for predicting for one user for a small subset of items.
        optimized for minimal latency for use in real-time ranking
        will return Nones for combinations of unknown user / unknown items

        :param user_id: a single user ID, may be an unknown users (all predictions will be None)
        :param item_ids: a subset of item IDs, may have unknown items (prediction for those will be None)
        :param rank_training_last:  if set to True predictions for interactions seen during training
            seen during will be ranked last by being set to -np.inf
        :param sort: whether to sort the result in decreasing order of prediction (best first)
        :return: a pandas DataFrame of userid | itemid | prediction,
            sorted in decresing order of prediction (if sort=True),
            with Nones for unknown users or items
        """

        df = pd.DataFrame()
        df[self._item_col] = item_ids  # assigning first because determines length
        df[self._user_col] = user_id
        df[self._prediction_col] = None

        new_users_mask = self.sparse_mat_builder.uid_encoder.find_new_labels([user_id])
        if np.any(new_users_mask):
            return df

        new_items_mask = self.sparse_mat_builder.iid_encoder.find_new_labels(item_ids)

        preds = self._predict_for_users_dense_direct(
            user_ids=[user_id],
            item_ids=np.array(item_ids)[~new_items_mask],
            exclude_training=rank_training_last)

        df[self._prediction_col].values[~new_items_mask] = preds.ravel()

        if sort:
            df.sort_values(self._prediction_col, ascending=False, inplace=True)

        return df[[self._user_col, self._item_col, self._prediction_col ]]  # reordering

